from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionCheckCardExistence(Action):
    def name(self) -> str:
        return "action_check_card_existence"

    def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[str, Any]
    ) -> List[Dict[Text, Any]]:
        username = tracker.get_slot("username")
        card_number = tracker.get_slot("card_number")

        db = Database()

        # Get user information
        user = db.get_user_by_name(username)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Get all cards for the user
        cards = db.get_cards_by_user(user["id"])
        card_numbers = [card["number"] for card in cards]

        if card_number in card_numbers:
            return [SlotSet("card_found", True)]
        else:
            return [SlotSet("card_found", False)]



from __future__ import annotations

import re
from typing import Any, Dict, List, Optional, Protocol, runtime_checkable

from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@runtime_checkable
class PromptCommand(Protocol):
    """
    A protocol for commands predicted by the LLM model and incorporated into the prompt.
    """

    @classmethod
    def command(cls) -> str:
        """
        Returns the command name.

        This class method should be implemented to return the name of the command.
        """
        ...

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> PromptCommand:
        """
        Converts the dictionary to a command.

        This class method should be implemented to create a command instance from the
        given dictionary.
        """
        ...

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """
        Runs the command on the tracker.

        This method should be implemented to execute the command on the given tracker.
        """
        ...

    def __hash__(self) -> int:
        """
        Returns the hash of the command.

        This method should be implemented to return the hash of the command.
        Useful for comparing commands and storing them in sets.
        """
        ...

    def __eq__(self, other: object) -> bool:
        """
        Compares the command with another object.

        This method should be implemented to compare the command with another object.
        """
        ...

    def to_dsl(self) -> str:
        """
        Converts the command to a DSL string.

        This method should be implemented to convert the command to a DSL string.
        A DSL string is a string representation of the command that is used in the
        prompt.
        """
        ...

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> Optional[PromptCommand]:
        """
        Converts the regex match to a command.

        This class method should be implemented to create a command instance from the
        given DSL string.
        """
        ...

    @staticmethod
    def regex_pattern() -> str:
        """
        Returns the regex pattern for the command.

        This method should be implemented to return the regex pattern that matches the
        command in the prompt.
        """
        ...



from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, Dict, List, Optional, Set, Text, Tuple, Union

import structlog
from jinja2 import Environment, Template, select_autoescape

import rasa.dialogue_understanding.generator.utils
import rasa.shared.utils.io
from rasa.dialogue_understanding.commands import (
    Command,
    SetSlotCommand,
    StartFlowCommand,
)
from rasa.dialogue_understanding.constants import KEY_MINIMIZE_NUM_CALLS
from rasa.dialogue_understanding.generator import CommandGenerator
from rasa.dialogue_understanding.generator._jinja_filters import to_json_escaped_string
from rasa.dialogue_understanding.generator.constants import (
    DEFAULT_LLM_CONFIG,
    FLOW_RETRIEVAL_ACTIVE_KEY,
    FLOW_RETRIEVAL_FLOW_THRESHOLD,
    FLOW_RETRIEVAL_KEY,
    LLM_CONFIG_KEY,
    TO_JSON_ESCAPED_STRING_JINJA_FILTER,
)
from rasa.dialogue_understanding.generator.flow_retrieval import FlowRetrieval
from rasa.dialogue_understanding.stack.utils import top_flow_frame
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.constants import SetSlotExtractor
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import Flow, FlowsList, FlowStep
from rasa.shared.core.flows.steps.collect import CollectInformationFlowStep
from rasa.shared.core.slot_mappings import SlotFillingManager
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import FileIOException, ProviderClientAPIException
from rasa.shared.nlu.constants import FLOWS_IN_PROMPT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.providers.llm.llm_response import LLMResponse, measure_llm_latency
from rasa.shared.utils.health_check.llm_health_check_mixin import LLMHealthCheckMixin
from rasa.shared.utils.llm import (
    allowed_values_for_slot,
    llm_factory,
    resolve_model_client_config,
)
from rasa.utils.log_utils import log_llm

structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class LLMBasedCommandGenerator(
    LLMHealthCheckMixin, GraphComponent, CommandGenerator, ABC
):
    """An abstract class defining interface and common functionality
    of an LLM-based command generators.
    """

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        **kwargs: Any,
    ) -> None:
        super().__init__(config)
        self.config = {**self.get_default_config(), **config}
        self.config[LLM_CONFIG_KEY] = resolve_model_client_config(
            self.config.get(LLM_CONFIG_KEY), LLMBasedCommandGenerator.__name__
        )
        self._model_storage = model_storage
        self._resource = resource
        self.flow_retrieval: Optional[FlowRetrieval]

        if self.enabled_flow_retrieval:
            self.flow_retrieval = FlowRetrieval(
                self.config[FLOW_RETRIEVAL_KEY], model_storage, resource
            )
            structlogger.info("llm_based_command_generator.flow_retrieval.enabled")
            self.config[FLOW_RETRIEVAL_KEY] = self.flow_retrieval.config
        else:
            self.flow_retrieval = None

    ### Abstract methods
    @staticmethod
    @abstractmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config."""
        pass

    @classmethod
    @abstractmethod
    def load(
        cls: Any,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "LLMBasedCommandGenerator":
        """Loads trained component (see parent class for full docstring)."""
        pass

    @abstractmethod
    def persist(self) -> None:
        """Persist the component to disk for future loading."""
        pass

    @abstractmethod
    async def predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        **kwargs: Any,
    ) -> List[Command]:
        """Predict commands using the LLM.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.
            **kwargs: Keyword arguments for forward compatibility.

        Returns:
            The commands generated by the llm.
        """
        pass

    @abstractmethod
    def parse_commands(
        cls, actions: Optional[str], tracker: DialogueStateTracker, flows: FlowsList
    ) -> List[Command]:
        """Parse the actions returned by the llm into intent and entities.

        Args:
            actions: The actions returned by the llm.
            tracker: The tracker containing the current state of the conversation.
            flows: the list of flows

        Returns:
            The parsed commands.
        """
        pass

    @classmethod
    @abstractmethod
    def fingerprint_addon(cls: Any, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint of the knowledge base for the graph."""
        pass

    ### Shared implementations of GraphComponent parent
    @classmethod
    def create(
        cls,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> "LLMBasedCommandGenerator":
        """Creates a new untrained component (see parent class for full docstring)."""
        return cls(config, model_storage, resource)

    def train(
        self, training_data: TrainingData, flows: FlowsList, domain: Domain
    ) -> Resource:
        """Train the llm based command generator. Stores all flows into a vector
        store.
        """
        self.perform_llm_health_check(
            self.config.get(LLM_CONFIG_KEY),
            self.get_default_llm_config(),
            "llm_based_command_generator.train",
            LLMBasedCommandGenerator.__name__,
        )

        if (
            self.flow_retrieval is None
            and len(flows.user_flows) > FLOW_RETRIEVAL_FLOW_THRESHOLD
        ):
            structlogger.warn(
                "llm_based_command_generator.flow_retrieval.disabled",
                event_info=(
                    f"You have {len(flows.user_flows)} user flows but flow "
                    f"retrieval is disabled. "
                    f"It is recommended to enable flow retrieval if the "
                    f"total number of user flows exceed "
                    f"{FLOW_RETRIEVAL_FLOW_THRESHOLD}. "
                    f"Keeping it disabled can result in deterioration of "
                    f"command generator's functional "
                    f"performance and higher costs because of increased "
                    f"number of tokens in the prompt. For more"
                    "information see:\n"
                    "https://rasa.com/docs/rasa-pro/concepts/dialogue-understanding#how-the-llmcommandgenerator-works"
                ),
            )

        # flow retrieval is populated with only user-defined flows
        try:
            if self.flow_retrieval is not None and not flows.is_empty():
                self.flow_retrieval.populate(flows.user_flows, domain)
        except Exception as e:
            structlogger.error(
                "llm_based_command_generator.train.failed",
                event_info="Flow retrieval store is inaccessible.",
                error=e,
            )
            raise

        self.persist()
        return self._resource

    ### Helper methods
    @property
    def enabled_flow_retrieval(self) -> bool:
        return self.config[FLOW_RETRIEVAL_KEY].get(FLOW_RETRIEVAL_ACTIVE_KEY, True)

    @lru_cache
    def compile_template(self, template: str) -> Template:
        """
        Compile the prompt template and register custom filters.

        Compiling the template is an expensive operation,
        so we cache the result.
        """
        # Create an environment
        # Autoescaping disabled explicitly for LLM prompt templates rendered from
        # strings (safe, not HTML)
        env = Environment(
            autoescape=select_autoescape(
                disabled_extensions=["jinja2"], default_for_string=False, default=True
            )
        )
        # Register filters
        env.filters[TO_JSON_ESCAPED_STRING_JINJA_FILTER] = to_json_escaped_string

        # Return the template which can leverage registered filters
        return env.from_string(template)

    @classmethod
    def load_prompt_template_from_model_storage(
        cls,
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template_file_name: str,
    ) -> Optional[Text]:
        try:
            with model_storage.read_from(resource) as path:
                return rasa.shared.utils.io.read_file(path / prompt_template_file_name)
        except (FileNotFoundError, FileIOException) as e:
            structlogger.warning(
                "llm_based_command_generator.load_prompt_template.failed",
                error=e,
                resource=resource.name,
            )
        return None

    @classmethod
    def load_flow_retrival(
        cls,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
    ) -> Optional[FlowRetrieval]:
        """Load the FlowRetrieval component if it is enabled in the configuration."""
        enable_flow_retrieval = config.get(FLOW_RETRIEVAL_KEY, {}).get(
            FLOW_RETRIEVAL_ACTIVE_KEY, True
        )
        if enable_flow_retrieval:
            return FlowRetrieval.load(
                config=config.get(FLOW_RETRIEVAL_KEY),
                model_storage=model_storage,
                resource=resource,
            )
        return None

    async def filter_flows(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
    ) -> FlowsList:
        """Filters the available flows.

        Args:
            message: The message from the user.
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.

        Returns:
            Filtered list of flows.
        """
        # If the flow retrieval is disabled, use the all the provided flows.
        filtered_flows = (
            await self.flow_retrieval.filter_flows(tracker, message, flows)
            if self.flow_retrieval is not None
            else flows
        )
        # Filter flows based on current context (tracker and message)
        # to identify which flows LLM can potentially start.
        if tracker:
            filtered_flows = tracker.get_startable_flows(filtered_flows)
        else:
            filtered_flows = filtered_flows

        # add the filtered flows to the message for evaluation purposes
        message.set(
            FLOWS_IN_PROMPT, list(filtered_flows.user_flow_ids), add_to_output=True
        )
        log_llm(
            logger=structlogger,
            log_module="LLMBasedCommandGenerator",
            log_event="llm_based_command_generator.predict_commands.filtered_flows",
            message=message.data,
            enabled_flow_retrieval=self.flow_retrieval is not None,
            relevant_flows=list(filtered_flows.user_flow_ids),
        )
        return filtered_flows

    @measure_llm_latency
    async def invoke_llm(
        self, prompt: Union[List[dict], List[str], str]
    ) -> Optional[LLMResponse]:
        """Use LLM to generate a response.

        Args:
            prompt: The prompt can be,
                - a list of preformatted messages. Each message should be a dictionary
                    with the following keys:
                    - content: The message content.
                    - role: The role of the message (e.g. user or system).
                - a list of messages. Each message is a string and will be formatted
                    as a user message.
                - a single message as a string which will be formatted as user message.

        Returns:
            An LLMResponse object.

        Raises:
            ProviderClientAPIException: If an error occurs during the LLM API call.
        """
        llm = llm_factory(
            self.config.get(LLM_CONFIG_KEY), self.get_default_llm_config()
        )
        try:
            return await llm.acompletion(prompt)
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error("llm_based_command_generator.llm.error", error=e)
            raise ProviderClientAPIException(
                message="LLM call exception", original_exception=e
            )

    def prepare_flows_for_template(
        self, flows: FlowsList, tracker: DialogueStateTracker
    ) -> List[Dict[str, Any]]:
        """Format data on available flows for insertion into the prompt template.

        Args:
            flows: The flows available to the user.
            tracker: The tracker containing the current state of the conversation.

        Returns:
            The inputs for the prompt template.
        """
        result = []
        for flow in flows.user_flows:
            slots_with_info = [
                {
                    "name": q.collect,
                    "description": q.description,
                    "allowed_values": allowed_values_for_slot(tracker.slots[q.collect]),
                }
                for q in flow.get_collect_steps()
                if self.is_extractable(q, tracker)
            ]
            result.append(
                {
                    "name": flow.id,
                    "description": flow.description,
                    "slots": slots_with_info,
                }
            )
        return result

    @staticmethod
    def is_extractable(
        collect_step: CollectInformationFlowStep,
        tracker: DialogueStateTracker,
        current_step: Optional[FlowStep] = None,
    ) -> bool:
        """Check if the `collect` can be filled.

        A collect slot can only be filled if the slot exist
        and either the collect has been asked already or the
        slot has been filled already.

        Args:
            collect_step: The collect_information step.
            tracker: The tracker containing the current state of the conversation.
            current_step: The current step in the flow.

        Returns:
            `True` if the slot can be filled, `False` otherwise.
        """
        slot = tracker.slots.get(collect_step.collect)
        if slot is None:
            return False

        return (
            # we can fill because this is a slot that can be filled ahead of time
            not collect_step.ask_before_filling
            # we can fill because the slot has been filled already
            or slot.has_been_set
            # we can fill because the is currently getting asked
            or (
                current_step is not None
                and isinstance(current_step, CollectInformationFlowStep)
                and current_step.collect == collect_step.collect
            )
        )

    @staticmethod
    def get_slot_value(tracker: DialogueStateTracker, slot_name: str) -> str:
        """Get the slot value from the tracker.

        Args:
            tracker: The tracker containing the current state of the conversation.
            slot_name: The name of the slot.

        Returns:
            The slot value as a string.
        """
        slot_value = tracker.get_slot(slot_name)
        if slot_value is None:
            return "undefined"
        else:
            return str(slot_value)

    def prepare_current_flow_slots_for_template(
        self, top_flow: Flow, current_step: FlowStep, tracker: DialogueStateTracker
    ) -> List[Dict[str, Any]]:
        """Prepare the current flow slots for the template.

        Args:
            top_flow: The top flow.
            current_step: The current step in the flow.
            tracker: The tracker containing the current state of the conversation.

        Returns:
            The slots with values, types, allowed values and a description.
        """
        if top_flow is not None:
            flow_slots = [
                {
                    "name": collect_step.collect,
                    "value": self.get_slot_value(tracker, collect_step.collect),
                    "type": tracker.slots[collect_step.collect].type_name,
                    "allowed_values": allowed_values_for_slot(
                        tracker.slots[collect_step.collect]
                    ),
                    "description": collect_step.description,
                }
                for collect_step in top_flow.get_collect_steps()
                if self.is_extractable(collect_step, tracker, current_step)
            ]
        else:
            flow_slots = []
        return flow_slots

    def prepare_current_slot_for_template(
        self, current_step: FlowStep
    ) -> Tuple[Union[str, None], Union[str, None]]:
        """Prepare the current slot for the template."""
        return (
            (current_step.collect, current_step.description)
            if isinstance(current_step, CollectInformationFlowStep)
            else (None, None)
        )

    @staticmethod
    def _prior_commands_contain_start_flow(prior_commands: List[Command]) -> bool:
        return any(isinstance(command, StartFlowCommand) for command in prior_commands)

    @staticmethod
    def _prior_commands_contain_set_slot_for_active_collect_step(
        prior_commands: List[Command],
        flows: FlowsList,
        tracker: DialogueStateTracker,
    ) -> bool:
        latest_user_frame = top_flow_frame(tracker.stack, ignore_call_frames=False)

        if latest_user_frame is None:
            return False

        active_flow = latest_user_frame.flow(flows)
        active_step = active_flow.step_by_id(latest_user_frame.step_id)

        if not isinstance(active_step, CollectInformationFlowStep):
            return False

        return any(
            command.name == active_step.collect
            for command in prior_commands
            if isinstance(command, SetSlotCommand)
        )

    def _should_skip_llm_call(
        self,
        prior_commands: List[Command],
        flows: FlowsList,
        tracker: DialogueStateTracker,
    ) -> bool:
        """Skip invoking the LLM.

        This returns True if the bot builder sets the property
        KEY_MINIMIZE_NUM_CALLS to True and the prior commands
        either contain a StartFlowCommand or a SetSlot command
        for the current collect step.
        """
        return self.config.get(KEY_MINIMIZE_NUM_CALLS, True) and (
            self._prior_commands_contain_start_flow(prior_commands)
            or self._prior_commands_contain_set_slot_for_active_collect_step(
                prior_commands, flows, tracker
            )
        )

    @staticmethod
    def _check_commands_against_slot_mappings(
        commands: List[Command],
        tracker: DialogueStateTracker,
        domain: Optional[Domain] = None,
    ) -> List[Command]:
        """Check if the LLM-issued slot commands are fillable.

        The LLM-issued slot commands are fillable if the slot
        mappings are satisfied (in particular the mapping conditions).
        """
        if not domain:
            return commands

        llm_fillable_slots = [
            tracker.slots.get(command.name)
            for command in commands
            if isinstance(command, SetSlotCommand)
            and command.extractor == SetSlotExtractor.LLM.value
            and tracker.slots.get(command.name) is not None
        ]

        if not llm_fillable_slots:
            return commands

        slot_filling_manager = SlotFillingManager(domain, tracker)
        slots_to_be_removed = []

        structlogger.debug(
            "command_processor.check_commands_against_slot_mappings.active_flow",
            active_flow=tracker.active_flow,
        )

        for slot in llm_fillable_slots:
            should_fill_slot = False
            for mapping in slot.mappings:  # type: ignore[union-attr]
                should_fill_slot = slot_filling_manager.should_fill_slot(
                    slot.name,  # type: ignore[union-attr]
                    mapping,
                )

                if should_fill_slot:
                    break

            if not should_fill_slot:
                structlogger.debug(
                    "command_processor.check_commands_against_slot_mappings.slot_not_fillable",
                    slot_name=slot.name,  # type: ignore[union-attr]
                )
                slots_to_be_removed.append(slot.name)  # type: ignore[union-attr]

        if not slots_to_be_removed:
            return commands

        filtered_commands = [
            command
            for command in commands
            if not (
                isinstance(command, SetSlotCommand)
                and command.name in slots_to_be_removed
            )
        ]

        return filtered_commands

    def _should_merge_llm_commands(
        self,
        prior_commands: List[Command],
        prior_start_flow_names: Set[str],
    ) -> bool:
        """Check if the LLM current commands should be merged with the prior commands.

        This can be done if there are no prior start flow commands.
        """
        return not prior_start_flow_names

    def _check_start_flow_command_overlap(
        self,
        prior_commands: List[Command],
        commands: List[Command],
        prior_start_flow_names: Set[str],
        current_start_flow_names: Set[str],
    ) -> List[Command]:
        """Prioritize the prior commands over the LLM-issued commands."""
        different_flow_names = current_start_flow_names.difference(
            prior_start_flow_names
        )

        if not different_flow_names or self._should_merge_llm_commands(
            prior_commands, prior_start_flow_names
        ):
            return prior_commands + commands

        # discard the flow names that are different to prior start flow commands
        filtered_commands = [
            command
            for command in commands
            if not isinstance(command, StartFlowCommand)
            or command.flow not in different_flow_names
        ]
        return prior_commands + filtered_commands

    def _filter_slot_commands(
        self,
        prior_commands: List[Command],
        commands: List[Command],
        overlapping_slot_names: Set[str],
    ) -> Tuple[List[Command], List[Command]]:
        """Prioritize prior commands over LLM ones in the case of same slot."""
        filtered_commands = (
            rasa.dialogue_understanding.generator.utils.filter_slot_commands(
                commands, overlapping_slot_names
            )
        )
        return prior_commands, filtered_commands

    @staticmethod
    def get_default_llm_config() -> Dict[str, Any]:
        """Get the default LLM config for the command generator."""
        return DEFAULT_LLM_CONFIG



import csv

from rasa_sdk import Action
from rasa_sdk.events import SlotSet


class ActionGetCustomerInfo(Action):
    def name(self):
        return "action_get_customer_info"

    def run(self, dispatcher, tracker, domain):
        # Load CSV file
        file_path = "csvs/customers.csv"  # get information from your DBs
        customer_id = tracker.get_slot("customer_id")

        try:
            with open(file_path, "r", newline="") as csvfile:
                reader = csv.DictReader(csvfile)

                # Filter data for the given customer ID
                customer_info = None
                for row in reader:
                    if row["customer_id"] == str(customer_id):
                        customer_info = row
                        break

                if customer_info is None:
                    dispatcher.utter_message("No customer found with this ID.")
                    return []

                # Extract customer details
                first_name = customer_info["first_name"]

                # Set the retrieved name in a slot
                return [SlotSet("customer_first_name", first_name)]

        except FileNotFoundError:
            dispatcher.utter_message("Customer database file not found.")
            return []
        except Exception:
            dispatcher.utter_message("Error retrieving customer information.")
            return []



from __future__ import annotations

from dataclasses import dataclass
from typing import Text

from rasa.shared.core.flows.flow_step_links import (
    FlowStepLinks,
    StaticFlowStepLink,
)
from rasa.shared.core.flows.steps.constants import START_STEP, UNSET_FLOW_STEP_ID
from rasa.shared.core.flows.steps.internal import InternalFlowStep


@dataclass
class StartFlowStep(InternalFlowStep):
    """A dynamically added flow step that represents the beginning of a flow."""

    def __init__(self, flow_id: Text, start_step_id: Text) -> None:
        """Initializes a start flow step.

        Args:
            flow_id: The id of the flow that contains the step
            start_step_id: The step id of the first step of the flow
        """
        super().__init__(
            idx=UNSET_FLOW_STEP_ID,
            custom_id=START_STEP,
            description=None,
            metadata={},
            next=FlowStepLinks(links=[StaticFlowStepLink(start_step_id)]),
            flow_id=flow_id,
        )



"""Functions for training and loading Rasa models."""

import tempfile

import structlog

from rasa.builder import config
from rasa.builder.exceptions import AgentLoadError, TrainingError
from rasa.core import agent
from rasa.core.utils import AvailableEndpoints, read_endpoints_from_path
from rasa.model_training import train
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string

structlogger = structlog.get_logger()


async def train_and_load_agent(importer: TrainingDataImporter) -> agent.Agent:
    """Train a model and load an agent.

    Args:
        importer: Training data importer with domain, flows, and config

    Returns:
        Loaded and ready agent

    Raises:
        TrainingError: If training fails
        AgentLoadError: If agent loading fails
    """
    try:
        # Setup endpoints for training validation
        await _setup_endpoints()

        # Train the model
        training_result = await _train_model(importer)

        # Load the agent
        agent_instance = await _load_agent(training_result.model)

        # Verify agent is ready
        if not agent_instance.is_ready():
            raise AgentLoadError("Agent failed to load properly - model is not ready")

        structlogger.info("training.agent_ready", model_path=training_result.model)

        return agent_instance

    except (TrainingError, AgentLoadError):
        raise
    except Exception as e:
        raise TrainingError(f"Unexpected error during training: {e}")


async def _setup_endpoints():
    """Setup endpoints configuration for training."""
    try:
        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".yml", delete=False
        ) as temp_file:
            endpoints_yaml = dump_obj_as_yaml_to_string(config.get_default_endpoints())
            temp_file.write(endpoints_yaml)
            temp_file.flush()

            # Reset and load endpoints
            AvailableEndpoints.reset_instance()
            read_endpoints_from_path(temp_file.name)

            structlogger.debug("training.endpoints_setup", temp_file=temp_file.name)

    except Exception as e:
        raise TrainingError(f"Failed to setup endpoints: {e}")


async def _train_model(importer: TrainingDataImporter):
    """Train the Rasa model."""
    try:
        structlogger.info("training.started")

        training_result = await train(
            domain="",
            config="",
            training_files=None,
            file_importer=importer,
        )

        if not training_result or not training_result.model:
            raise TrainingError("Training completed but no model was produced")

        structlogger.info("training.completed", model_path=training_result.model)

        return training_result

    except Exception as e:
        raise TrainingError(f"Model training failed: {e}")


async def _load_agent(model_path: str) -> agent.Agent:
    """Load the trained agent."""
    try:
        structlogger.info("training.loading_agent", model_path=model_path)

        available_endpoints = AvailableEndpoints.get_instance()
        if available_endpoints is None:
            raise AgentLoadError("No endpoints available for agent loading")

        agent_instance = await agent.load_agent(
            model_path=model_path,
            remote_storage=None,
            endpoints=available_endpoints,
        )

        if agent_instance is None:
            raise AgentLoadError("Agent loading returned None")

        structlogger.info("training.agent_loaded", model_path=model_path)

        return agent_instance

    except AgentLoadError:
        raise
    except Exception as e:
        raise AgentLoadError(f"Failed to load agent: {e}")



import json
import logging
import os
import typing
from typing import Any, Callable, Dict, List, Optional, Text

import rasa.shared.utils.io
from rasa.shared.nlu.training_data.formats.dialogflow import (
    DIALOGFLOW_AGENT,
    DIALOGFLOW_ENTITIES,
    DIALOGFLOW_ENTITY_ENTRIES,
    DIALOGFLOW_INTENT,
    DIALOGFLOW_INTENT_EXAMPLES,
    DIALOGFLOW_PACKAGE,
)
from rasa.shared.nlu.training_data.training_data import TrainingData

if typing.TYPE_CHECKING:
    from rasa.shared.nlu.training_data.formats.readerwriter import TrainingDataReader

logger = logging.getLogger(__name__)

# Different supported file formats and their identifier
WIT = "wit"
LUIS = "luis"
RASA = "rasa_nlu"
RASA_YAML = "rasa_yml"
UNK = "unk"
DIALOGFLOW_RELEVANT = {DIALOGFLOW_ENTITIES, DIALOGFLOW_INTENT}

_json_format_heuristics: Dict[Text, Callable[[Any, Text], bool]] = {
    WIT: lambda js, fn: "utterances" in js and "luis_schema_version" not in js,
    LUIS: lambda js, fn: "luis_schema_version" in js,
    RASA: lambda js, fn: "rasa_nlu_data" in js,
    DIALOGFLOW_AGENT: lambda js, fn: "supportedLanguages" in js,
    DIALOGFLOW_PACKAGE: lambda js, fn: "version" in js and len(js) == 1,
    DIALOGFLOW_INTENT: lambda js, fn: "responses" in js,
    DIALOGFLOW_ENTITIES: lambda js, fn: "isEnum" in js,
    DIALOGFLOW_INTENT_EXAMPLES: lambda js, fn: "_usersays_" in fn,
    DIALOGFLOW_ENTITY_ENTRIES: lambda js, fn: "_entries_" in fn,
}


def load_data(resource_name: Text, language: Optional[Text] = "en") -> "TrainingData":
    """Load training data from disk.

    Merges them if loaded from disk and multiple files are found.
    """
    if not os.path.exists(resource_name):
        raise ValueError(f"File '{resource_name}' does not exist.")

    if os.path.isfile(resource_name):
        files = [resource_name]
    else:
        files = rasa.shared.utils.io.list_files(resource_name)

    data_sets = [_load(f, language) for f in files]
    training_data_sets: List[TrainingData] = [ds for ds in data_sets if ds]
    if len(training_data_sets) == 0:
        training_data = TrainingData()
    elif len(training_data_sets) == 1:
        training_data = training_data_sets[0]
    else:
        training_data = training_data_sets[0].merge(*training_data_sets[1:])

    return training_data


def _reader_factory(fformat: Text) -> Optional["TrainingDataReader"]:
    """Generates the appropriate reader class based on the file format."""
    from rasa.shared.nlu.training_data.formats import (
        DialogflowReader,
        LuisReader,
        RasaReader,
        RasaYAMLReader,
        WitReader,
    )

    reader: Optional["TrainingDataReader"] = None
    if fformat == LUIS:
        reader = LuisReader()
    elif fformat == WIT:
        reader = WitReader()
    elif fformat in DIALOGFLOW_RELEVANT:
        reader = DialogflowReader()
    elif fformat == RASA:
        reader = RasaReader()
    elif fformat == RASA_YAML:
        reader = RasaYAMLReader()
    return reader


def _load(filename: Text, language: Optional[Text] = "en") -> Optional["TrainingData"]:
    """Loads a single training data file from disk."""
    fformat = guess_format(filename)
    if fformat == UNK:
        raise ValueError(f"Unknown data format for file '{filename}'.")

    reader = _reader_factory(fformat)

    if reader:
        return reader.read(filename, language=language, fformat=fformat)
    else:
        return None


def guess_format(filename: Text) -> Text:
    """Applies heuristics to guess the data format of a file.

    Args:
        filename: file whose type should be guessed

    Returns:
        Guessed file format.
    """
    from rasa.shared.nlu.training_data.formats import RasaYAMLReader

    guess = UNK

    if not os.path.isfile(filename):
        return guess

    try:
        content = rasa.shared.utils.io.read_file(filename)
        js = json.loads(content)
    except ValueError:
        if RasaYAMLReader.is_yaml_nlu_file(filename):
            guess = RASA_YAML
    else:
        for file_format, format_heuristic in _json_format_heuristics.items():
            if format_heuristic(js, filename):
                guess = file_format
                break

    logger.debug(f"Training data format of '{filename}' is '{guess}'.")

    return guess



from dataclasses import dataclass
from typing import Optional

import structlog

from rasa.shared.exceptions import InvalidConfigException

structlogger = structlog.get_logger()


def validate_username_password_credentials(
    username: Optional[str], password: Optional[str], channel_name: str
) -> None:
    """Validate that username and password are both provided or both None.

    Args:
        username: The username credential, or None
        password: The password credential, or None
        channel_name: The name of the channel for error message

    Raises:
        InvalidConfigException: If only one of username/password is provided
    """
    if (not username) != (not password):
        raise InvalidConfigException(
            f"In {channel_name} channel, either both username and password "
            "or neither should be provided."
        )


def validate_voice_license_scope() -> None:
    from rasa.utils.licensing import (
        PRODUCT_AREA,
        VOICE_SCOPE,
        validate_license_from_env,
    )

    """Validate that the correct license scope is present."""
    structlogger.info(
        f"Validating current Rasa Pro license scope which must include "
        f"the '{VOICE_SCOPE}' scope to use the voice channel."
    )

    voice_product_scope = PRODUCT_AREA + " " + VOICE_SCOPE
    validate_license_from_env(product_area=voice_product_scope)


@dataclass
class CallParameters:
    """Standardized call parameters for voice channels."""

    call_id: str
    user_phone: str
    bot_phone: Optional[str] = None
    user_name: Optional[str] = None
    user_host: Optional[str] = None
    bot_host: Optional[str] = None
    direction: Optional[str] = None
    stream_id: Optional[str] = None



import logging
from typing import Dict, Optional, Text

from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.training_data.structures import StoryGraph
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.utils.common import cached_method

logger = logging.getLogger(__name__)


class StaticTrainingDataImporter(TrainingDataImporter):
    """Static `TrainingFileImporter` implementation."""

    def __init__(
        self,
        domain: Domain,
        stories: Optional[StoryGraph] = None,
        flows: Optional[FlowsList] = None,
        nlu_data: Optional[TrainingData] = None,
        config: Optional[Dict] = None,
    ):
        self.domain = domain
        self.stories = stories or StoryGraph([])
        self.flows = flows or FlowsList(underlying_flows=[])
        self.nlu_data = nlu_data or TrainingData()
        self.config = config or {}

    @cached_method
    def get_config(self) -> Dict:
        """Retrieves model config (see parent class for full docstring)."""
        return self.config

    def get_config_file_for_auto_config(self) -> Optional[Text]:
        """Returns config file path for auto-config only if there is a single one."""
        return None

    @cached_method
    def get_stories(self, exclusion_percentage: Optional[int] = None) -> StoryGraph:
        """Retrieves training stories / rules (see parent class for full docstring)."""
        return self.stories

    @cached_method
    def get_flows(self) -> FlowsList:
        """Retrieves training stories / rules (see parent class for full docstring)."""
        return self.flows

    @cached_method
    def get_conversation_tests(self) -> StoryGraph:
        """Retrieves conversation test stories (see parent class for full docstring)."""
        return StoryGraph([])

    @cached_method
    def get_nlu_data(self, language: Optional[Text] = "en") -> TrainingData:
        """Retrieves NLU training data (see parent class for full docstring)."""
        return self.nlu_data

    @cached_method
    def get_domain(self) -> Domain:
        """Retrieves model domain (see parent class for full docstring)."""
        return self.domain



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    HUGGINGFACE_CACHE_FOLDER_CONFIG_KEY,
    HUGGINGFACE_ENCODE_KWARGS_CONFIG_KEY,
    HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER,
    HUGGINGFACE_MODEL_KWARGS_CONFIG_KEY,
    HUGGINGFACE_MULTIPROCESS_CONFIG_KEY,
    HUGGINGFACE_SHOW_PROGRESS_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    RASA_TYPE_CONFIG_KEY,
    REQUEST_TIMEOUT_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.utils.configs import (
    raise_deprecation_warnings,
    resolve_aliases,
    validate_required_keys,
)
from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()

DEPRECATED_HUGGINGFACE_TYPE = "huggingface"

DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING = {
    # Provider aliases
    RASA_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    # Model name aliases
    MODEL_NAME_CONFIG_KEY: MODEL_CONFIG_KEY,
    # Timeout aliases
    REQUEST_TIMEOUT_CONFIG_KEY: TIMEOUT_CONFIG_KEY,
}

REQUIRED_KEYS = [MODEL_CONFIG_KEY, PROVIDER_CONFIG_KEY]


@dataclass
class HuggingFaceLocalEmbeddingClientConfig:
    """Parses configuration for HuggingFace local embeddings client, resolves
    aliases and raises deprecation warnings.

    Raises:
        ValueError: Raised in cases of invalid configuration:
            - If any of the required configuration keys are missing.
            - If `api_type` has a value different from `huggingface_local` or
              `huggingface` (deprecated).
    """

    model: str

    multi_process: Optional[bool]
    cache_folder: Optional[str]
    show_progress: Optional[bool]

    # Provider is not actually used by sentence-transformers, but we define
    # it here because it's used as a switch denominator for HuggingFace
    # local embedding client.
    provider: str = HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER

    model_kwargs: dict = field(default_factory=dict)
    encode_kwargs: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.provider != HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER:
            message = (
                f"API type must be set to '{HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER}'."
            )
            structlogger.error(
                "huggingface_local_embeddings_client_config.validation_error",
                message=message,
                provider=self.provider,
            )
            raise ValueError(message)
        if self.model is None:
            message = "Model cannot be set to None."
            structlogger.error(
                "huggingface_local_embeddings_client_config.validation_error",
                message=message,
                model=self.model,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> HuggingFaceLocalEmbeddingClientConfig:
        """
        Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            DefaultLiteLLMClientConfig
        """
        # Check for usage of deprecated switching key and value:
        # 1. type: huggingface
        # 2. _type: huggingface
        _raise_deprecation_warning_for_huggingface_deprecated_switch_value(config)
        # Check for other deprecated keys
        raise_deprecation_warnings(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)
        # Resolve any potential aliases
        config = cls.resolve_config_aliases(config)
        # Validate that required keys are set
        validate_required_keys(config, REQUIRED_KEYS)
        this = HuggingFaceLocalEmbeddingClientConfig(
            # Required parameters
            model=config.pop(MODEL_CONFIG_KEY),
            provider=config.pop(PROVIDER_CONFIG_KEY),
            # Optional
            multi_process=config.pop(HUGGINGFACE_MULTIPROCESS_CONFIG_KEY, False),
            cache_folder=config.pop(
                HUGGINGFACE_CACHE_FOLDER_CONFIG_KEY,
                str(HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER),
            ),
            show_progress=config.pop(HUGGINGFACE_SHOW_PROGRESS_CONFIG_KEY, False),
            model_kwargs=config.pop(HUGGINGFACE_MODEL_KWARGS_CONFIG_KEY, {}),
            encode_kwargs=config.pop(HUGGINGFACE_ENCODE_KWARGS_CONFIG_KEY, {}),
        )
        return this

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        return asdict(self)

    @staticmethod
    def resolve_config_aliases(config: Dict[str, Any]) -> Dict[str, Any]:
        config = _resolve_huggingface_deprecated_switch_value(config)
        return resolve_aliases(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)


def is_huggingface_local_config(config: dict) -> bool:
    """Check whether the configuration is meant to configure
    a local HuggingFace embedding client.
    """
    # Hugging face special deprecated cases:
    # 1. type: huggingface
    # 2. _type: huggingface
    # If the deprecated setting is detected resolve both alias key and key
    # value. This would mean that the configurations above will be
    # transformed to:
    # provider: huggingface_local
    config = HuggingFaceLocalEmbeddingClientConfig.resolve_config_aliases(config)

    # Case: Configuration contains `provider: huggingface_local`
    if config.get(PROVIDER_CONFIG_KEY) in [
        HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER,
    ]:
        return True

    return False


def _raise_deprecation_warning_for_huggingface_deprecated_switch_value(
    config: dict,
) -> None:
    deprecated_switch_keys = [RASA_TYPE_CONFIG_KEY, LANGCHAIN_TYPE_CONFIG_KEY]
    deprecation_message = (
        f"Configuration "
        f"`{{deprecated_switch_key}}: {DEPRECATED_HUGGINGFACE_TYPE}` "
        f"is deprecated and will be removed in 4.0.0. "
        f"Please use "
        f"`{PROVIDER_CONFIG_KEY}: {HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER}` "
        f"instead."
    )
    for deprecated_switch_key in deprecated_switch_keys:
        if (
            deprecated_switch_key in config
            and config[deprecated_switch_key] == DEPRECATED_HUGGINGFACE_TYPE
        ):
            raise_deprecation_warning(
                message=deprecation_message.format(
                    deprecated_switch_key=deprecated_switch_key
                )
            )


def _resolve_huggingface_deprecated_switch_value(config: dict) -> dict:
    """
    Resolve use of deprecated switching mechanism for HuggingFace local
    embedding client.

    The following settings (key + value) are deprecated:
    1. `type: huggingface`
    2. `_type: huggingface`
    in favor of `provider: huggingface_local`.


    Args:
        config: given config

    Returns:
        New config with resolved switch mechanism

    """
    config = config.copy()

    deprecated_switch_keys = [RASA_TYPE_CONFIG_KEY, LANGCHAIN_TYPE_CONFIG_KEY]
    debug_message = (
        f"Switching "
        f"`{{deprecated_switch_key}}: {DEPRECATED_HUGGINGFACE_TYPE}` "
        f"to `{PROVIDER_CONFIG_KEY}: {HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER}`."
    )

    for deprecated_switch_key in deprecated_switch_keys:
        if (
            deprecated_switch_key in config
            and config[deprecated_switch_key] == DEPRECATED_HUGGINGFACE_TYPE
        ):
            # Update configuration with new switch mechanism
            config[PROVIDER_CONFIG_KEY] = HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER
            # Pop the deprecated key used
            config.pop(deprecated_switch_key, None)

            structlogger.debug(
                "HuggingFaceLocalEmbeddingClientConfig"
                "._resolve_huggingface_deprecated_switch_value",
                message=debug_message.format(
                    deprecated_switch_key=deprecated_switch_key
                ),
                new_config=config,
            )

    return config



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.commands.utils import (
    clean_extracted_value,
    find_default_flows_collecting_slot,
    get_nullable_slot_value,
)
from rasa.dialogue_understanding.patterns.collect_information import (
    CollectInformationPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import (
    get_collect_steps_excluding_ask_before_filling_for_active_flow,
)
from rasa.shared.constants import ROUTE_TO_CALM_SLOT
from rasa.shared.core.constants import SetSlotExtractor
from rasa.shared.core.events import Event, SlotSet
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import SET_SLOT_COMMAND

structlogger = structlog.get_logger()


def get_flows_predicted_to_start_from_tracker(
    tracker: DialogueStateTracker,
) -> List[str]:
    """Returns the flows that are predicted to start from the current state.

    Args:
        tracker: The tracker to use.

    Returns:
        The flows that are predicted to start from the current state.
    """
    from rasa.dialogue_understanding.processor.command_processor import (
        filter_start_flow_commands,
        get_commands_from_tracker,
    )

    commands = get_commands_from_tracker(tracker)
    return filter_start_flow_commands(commands)


@dataclass
class SetSlotCommand(Command):
    """A command to set a slot."""

    name: str
    value: Any
    extractor: str = SetSlotExtractor.LLM.value

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return SET_SLOT_COMMAND

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> SetSlotCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return SetSlotCommand(
                name=data["name"],
                value=data["value"],
                extractor=data.get("extractor", SetSlotExtractor.LLM.value),
            )
        except KeyError as e:
            raise ValueError(f"Missing key when parsing SetSlotCommand: {e}") from e

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        slot = tracker.slots.get(self.name)
        if slot is None:
            structlogger.debug(
                "set_slot_command.skip_command.slot_not_in_domain", command=self
            )
            return []

        if slot.has_same_coerced_value(self.value):
            # value hasn't changed, skip this one
            structlogger.debug(
                "set_slot_command.skip_command.slot_already_set", command=self
            )
            return []

        # Get slots of the active flow
        slots_of_active_flow = (
            get_collect_steps_excluding_ask_before_filling_for_active_flow(
                tracker.stack, all_flows
            )
        )

        # Add slots that are asked in the current collect step. This is needed
        # to include slots that has ask_before_filling set to True.
        top_frame = tracker.stack.top()
        if isinstance(top_frame, CollectInformationPatternFlowStackFrame):
            slots_of_active_flow.add(top_frame.collect)

        if (
            self.name not in slots_of_active_flow
            and self.name != ROUTE_TO_CALM_SLOT
            and not slot.is_builtin
            and self.extractor
            in {
                SetSlotExtractor.LLM.value,
                SetSlotExtractor.COMMAND_PAYLOAD_READER.value,
                SetSlotExtractor.NLU.value,
            }
        ):
            # Get the other predicted flows from the most recent message on the tracker.
            predicted_flows = get_flows_predicted_to_start_from_tracker(tracker)
            if not predicted_flows:
                # If no predicted flows, check for default flows collecting the slot.
                predicted_flows = find_default_flows_collecting_slot(
                    self.name, all_flows
                )
            use_slot_fill = any(
                step.collect == self.name and not step.ask_before_filling
                for flow in all_flows.underlying_flows
                if flow.id in predicted_flows
                for step in flow.get_collect_steps()
            )
            if not use_slot_fill:
                structlogger.debug(
                    "set_slot_command.skip_command.slot_not_asked_for", command=self
                )
                return []

        structlogger.debug("set_slot_command.set_slot", command=self)
        return [
            SlotSet(self.name, slot.coerce_value(self.value), filled_by=self.extractor)
        ]

    def __hash__(self) -> int:
        return hash(self.value) + hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SetSlotCommand):
            return False

        return (
            str(other.value).lower() == str(self.value).lower()
            and other.name == self.name
        )

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: f"SetSlot({self.name}, {self.value})",
            CommandSyntaxVersion.v2: f"set slot {self.name} {self.value}",
            CommandSyntaxVersion.v3: f"set slot {self.name} {self.value}",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> SetSlotCommand:
        """Converts the DSL string to a command."""
        slot_name = str(match.group(1).strip())
        slot_value = clean_extracted_value(match.group(2))
        return SetSlotCommand(name=slot_name, value=get_nullable_slot_value(slot_value))

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: (
                r"""SetSlot\(['"]?([a-zA-Z_][a-zA-Z0-9_-]*)['"]?, ?['"]?(.*)['"]?\)"""
            ),
            CommandSyntaxVersion.v2: (
                r"""^[\s\W\d]*set slot ['"`]?([a-zA-Z_][a-zA-Z0-9_-]*)['"`]? ['"`]?(.+?)[\W]*$"""  # noqa: E501
            ),
            CommandSyntaxVersion.v3: (
                r"""^[\s\W\d]*set slot ['"`]?([a-zA-Z_][a-zA-Z0-9_-]*)['"`]? ['"`]?(.+?)[\W]*$"""  # noqa: E501
            ),
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



from __future__ import annotations

import csv
from pathlib import Path
from typing import Dict, List, Text, Tuple, Union

import numpy as np

from rasa.core.evaluation.marker_base import EventMetaData
from rasa.utils.io import WriteRow


def compute_statistics(
    values: List[Union[float, int]],
) -> Dict[Text, Union[int, float, np.floating]]:
    """Computes some statistics over the given numbers."""
    return {
        "count": len(values) if values else 0,
        "mean": np.mean(values) if values else np.nan,
        "median": np.median(values) if values else np.nan,
        "min": min(values) if values else np.nan,
        "max": max(values) if values else np.nan,
    }


class MarkerStatistics:
    """Computes some statistics on marker extraction results.

    (1) Number of sessions where markers apply:

    For each marker, we compute the total number (as well as the percentage) of
    sessions in which a marker applies at least once.
    Moreover, we output the total number of sessions that were parsed.

    (2) Number of user turns preceding relevant events - per sessions:

    For each marker, we consider all relevant events where that marker applies.
    Everytime a marker applies, we check how many user turns precede that event.
    We collect all these numbers and compute basic statistics (e.g. count and mean)
    on them.

    This means, per session, we compute how often a marker applies and how many
    user turns precede a relevant marker application on average, in that session.

    (3) Number of user turns preceding relevant events - over all sessions:

    Here, for each marker, we consider all relevant events where a marker applies
    *in any of the sessions*. Then, we again calculate basic statistics over the
    respective number of user turns that precede each of these events.

    This means, we compute how many events the marker applies in total and we
    compute an estimate of the expected number of user turns preceding that
    precede an (relevant) event where a marker applies.
    """

    NO_MARKER = "-"
    STAT_NUM_SESSIONS = "total_number_of_sessions"
    STAT_NUM_SESSIONS_WHERE_APPLIES = (
        "number_of_sessions_where_marker_applied_at_least_once"
    )
    STAT_PERCENTAGE_SESSIONS_WHERE_APPLIES = (
        "percentage_of_sessions_where_marker_applied_at_least_once"
    )

    @staticmethod
    def _add_num_user_turns_str_to(stat_name: Text) -> Text:
        return f"{stat_name}(number of preceding user turns)"

    ALL_SESSIONS = np.nan
    ALL_SENDERS = "all"

    def __init__(self) -> None:
        """Creates a new marker statistics object."""
        # to ensure consistency of processed rows
        self._marker_names: List[Text] = []

        # (1) For collecting the per-session analysis:
        # NOTE: we could stream / compute them later instead of collecting them...
        self.session_results: Dict[Text, Dict[Text, List[Union[int, float]]]] = {}
        self.session_identifier: List[Tuple[Text, int]] = []

        # (2) For the overall statistics:
        self.num_preceding_user_turns_collected: Dict[Text, List[int]] = {}
        self.count_if_applied_at_least_once: Dict[Text, int] = {}
        self.num_sessions = 0

    def process(
        self,
        sender_id: Text,
        session_idx: int,
        meta_data_on_relevant_events_per_marker: Dict[Text, List[EventMetaData]],
    ) -> None:
        """Processes the meta data that was extracted from a single session.

        Internally, this method ..
        1. computes some statistics for the given meta data and saves it for later
        2. keeps track of the total number of sessions processed and the
           collects all metadata to be able to compute meta data over *all*

        Args:
            sender_id: an id that, together with the `session_idx` identifies
                the session from which the markers where extracted
            session_idx: an index that, together with the `sender_id` identifies
                the session from which the markers where extracted
            meta_data_on_relevant_events_per_marker: marker extraction results,
                i.e. a dictionary mapping
                marker names to the meta data describing relevant events
                for those markers
        """
        if len(self._marker_names) == 0:
            # sort and initialise here once so our result tables are sorted
            self._marker_names = sorted(meta_data_on_relevant_events_per_marker.keys())
            self.count_if_applied_at_least_once = {
                marker_name: 0 for marker_name in self._marker_names
            }
            self.num_preceding_user_turns_collected = {
                marker_name: [] for marker_name in self._marker_names
            }
            # NOTE: we could stream / compute them later instead of collecting them...
            stat_names = sorted(compute_statistics([]).keys())
            self.session_results = {
                marker_name: {stat_name: [] for stat_name in stat_names}
                for marker_name in self._marker_names
            }
        else:
            given_markers = meta_data_on_relevant_events_per_marker.keys()
            if set(given_markers) != set(self._marker_names):
                raise RuntimeError(
                    f"Expected all processed extraction results to contain information"
                    f"for the same set of markers. But found "
                    f"{sorted(given_markers)} which differs from "
                    f"the marker extracted so far (i.e. {sorted(self._marker_names)})."
                )

        # update session identifiers / count
        self.num_sessions += 1
        self.session_identifier.append((sender_id, session_idx))

        for marker_name, meta_data in meta_data_on_relevant_events_per_marker.items():
            num_preceding_user_turns = [
                event_meta_data.preceding_user_turns for event_meta_data in meta_data
            ]

            # update per session statistics
            statistics = compute_statistics(num_preceding_user_turns)
            for stat_name, stat_value in statistics.items():
                self.session_results[marker_name][stat_name].append(stat_value)

            # update overall statistics
            self.num_preceding_user_turns_collected[marker_name].extend(
                num_preceding_user_turns
            )
            if len(num_preceding_user_turns):
                self.count_if_applied_at_least_once[marker_name] += 1

    def overall_statistic_to_csv(self, path: Path, overwrite: bool = False) -> None:
        """Exports the overall statistics (over all processes sessions) to a csv file.

        Args:
            path: path to where the csv file should be written.
            overwrite: set to `True` to enable overwriting an existing file
        """
        if path.is_file() and not overwrite:
            raise FileExistsError(f"Expected that there was no file at {path}.")
        with path.open(mode="w") as f:
            table_writer = csv.writer(f)
            table_writer.writerow(self._header())
            self._write_overview(table_writer)
            self._write_overall_statistics(table_writer)

    def per_session_statistics_to_csv(
        self, path: Path, overwrite: bool = False
    ) -> None:
        """Exports the resulting statistics to a csv file.

        Args:
            path: path to where the csv file should be written.
            overwrite: set to `True` to enable overwriting an existing file
        """
        if path.is_file() and not overwrite:
            raise FileExistsError(f"Expected that there was no file at {path}.")
        with path.open(mode="w") as f:
            table_writer = csv.writer(f)
            table_writer.writerow(self._header())
            # NOTE: we could stream / compute them later instead of collecting them...
            self._write_per_session_statistics(table_writer)

    @staticmethod
    def _header() -> List[Text]:
        return ["sender_id", "session_idx", "marker", "statistic", "value"]

    def _write_overview(self, table_writer: WriteRow) -> None:
        special_sender_idx = self.ALL_SENDERS
        special_session_idx = self.ALL_SESSIONS
        self._write_row(
            table_writer=table_writer,
            sender_id=special_sender_idx,
            session_idx=special_session_idx,
            marker_name=self.NO_MARKER,
            statistic_name=self.STAT_NUM_SESSIONS,
            statistic_value=self.num_sessions,
        )
        for marker_name, count in self.count_if_applied_at_least_once.items():
            self._write_row(
                table_writer=table_writer,
                sender_id=special_sender_idx,
                session_idx=special_session_idx,
                marker_name=marker_name,
                statistic_name=self.STAT_NUM_SESSIONS_WHERE_APPLIES,
                statistic_value=count,
            )
            self._write_row(
                table_writer=table_writer,
                sender_id=special_sender_idx,
                session_idx=special_session_idx,
                marker_name=marker_name,
                statistic_name=self.STAT_PERCENTAGE_SESSIONS_WHERE_APPLIES,
                statistic_value=(count / self.num_sessions * 100)
                if self.num_sessions
                else 100.0,
            )

    def _write_overall_statistics(self, table_writer: WriteRow) -> None:
        for marker_name, num_list in self.num_preceding_user_turns_collected.items():
            for statistic_name, value in compute_statistics(num_list).items():
                MarkerStatistics._write_row(
                    table_writer=table_writer,
                    sender_id=self.ALL_SENDERS,
                    session_idx=self.ALL_SESSIONS,
                    marker_name=marker_name,
                    statistic_name=self._add_num_user_turns_str_to(statistic_name),
                    statistic_value=value,
                )

    def _write_per_session_statistics(self, table_writer: WriteRow) -> None:
        for marker_name, collected_statistics in self.session_results.items():
            for statistic_name, values in collected_statistics.items():
                self._write_per_session_statistic(
                    table_writer=table_writer,
                    marker_name=marker_name,
                    statistic_name=statistic_name,
                    session_identifiers=self.session_identifier,
                    values=values,
                )

    @staticmethod
    def _write_per_session_statistic(
        table_writer: WriteRow,
        marker_name: Text,
        statistic_name: Text,
        session_identifiers: List[Tuple[Text, int]],
        values: List[Union[float, int]],
    ) -> None:
        for record_idx, (sender_id, session_idx) in enumerate(session_identifiers):
            MarkerStatistics._write_row(
                table_writer=table_writer,
                sender_id=sender_id,
                session_idx=session_idx,
                marker_name=marker_name,
                statistic_name=MarkerStatistics._add_num_user_turns_str_to(
                    statistic_name
                ),
                statistic_value=values[record_idx],
            )

    @staticmethod
    def _write_row(
        table_writer: WriteRow,
        sender_id: Text,
        session_idx: Union[int, float],
        marker_name: Text,
        statistic_name: Text,
        statistic_value: Union[int, float],
    ) -> None:
        if isinstance(statistic_value, int):
            value_str = str(statistic_value)
        elif np.isnan(statistic_value):
            value_str = str(np.nan)
        else:
            value_str = np.round(statistic_value, 3)
        table_writer.writerow(
            [
                str(item)
                for item in [
                    sender_id,
                    session_idx,
                    marker_name,
                    statistic_name,
                    value_str,
                ]
            ]
        )



import asyncio
import json
import logging
import os
import ssl
from asyncio import AbstractEventLoop
from collections import deque
from functools import cached_property
from typing import Any, Deque, Dict, List, Optional, Set, Text, Tuple, Union, cast
from urllib.parse import urlparse

import aio_pika
import structlog

import rasa.shared.utils.common
import rasa.shared.utils.io
from rasa.core.brokers.broker import EventBroker
from rasa.shared.constants import DOCS_URL_PIKA_EVENT_BROKER
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.io import DEFAULT_ENCODING
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

RABBITMQ_EXCHANGE = "rasa-exchange"
DEFAULT_QUEUE_NAME = "rasa_core_events"


class PikaEventBroker(EventBroker):
    """Pika-based event broker for publishing messages to RabbitMQ."""

    def __init__(
        self,
        host: Text,
        username: Text,
        password: Text,
        port: Union[int, Text] = 5672,
        queues: Union[List[Text], Tuple[Text, ...], Text, None] = None,
        should_keep_unpublished_messages: bool = True,
        raise_on_failure: bool = False,
        event_loop: Optional[AbstractEventLoop] = None,
        connection_attempts: int = 20,
        retry_delay_in_seconds: float = 5,
        exchange_name: Text = RABBITMQ_EXCHANGE,
        **kwargs: Any,
    ):
        """Initialise RabbitMQ event broker.

        Args:
            host: Pika host.
            username: Username for authentication with Pika host.
            password: Password for authentication with Pika host.
            port: port of the Pika host.
            queues: Pika queues to declare and publish to.
            should_keep_unpublished_messages: Whether or not the event broker should
                maintain a queue of unpublished messages to be published later in
                case of errors.
            raise_on_failure: Whether to raise an exception if publishing fails. If
                `False`, keep retrying.
            event_loop: The event loop which will be used to run `async` functions. If
                `None` `asyncio.get_event_loop()` is used to get a loop.
            connection_attempts: Number of attempts for connecting to RabbitMQ before
                an exception is thrown.
            retry_delay_in_seconds: Time in seconds between connection attempts.
            exchange_name: Exchange name to which the queues binds to.
                If nothing is mentioned then the default exchange name would be used.
        """
        super().__init__()

        self.host = host
        self.username = username
        self.password = password

        try:
            self.port = int(port)
        except ValueError as e:
            raise RasaException("Port could not be converted to integer.") from e

        self.queues = self._get_queues_from_args(queues)
        self.raise_on_failure = raise_on_failure
        self._connection_attempts = connection_attempts
        self._retry_delay_in_seconds = retry_delay_in_seconds
        self.exchange_name = exchange_name

        # Unpublished messages which hopefully will be published later 🤞
        self._unpublished_events: Deque[Dict[Text, Any]] = deque()
        self.should_keep_unpublished_messages = should_keep_unpublished_messages

        self._loop = event_loop or asyncio.get_event_loop()
        self._background_tasks: Set[asyncio.Task] = set()

        self._connection: Optional[aio_pika.abc.AbstractRobustConnection] = None
        self._exchange: Optional[aio_pika.RobustExchange] = None

        # PII related attributes
        self.stream_pii = kwargs.get("stream_pii", True)
        self.anonymization_queues = kwargs.get("anonymization_queues", [])

    @staticmethod
    def _get_queues_from_args(
        queues_arg: Union[List[Text], Tuple[Text, ...], Text, None],
    ) -> Union[List[Text], Tuple[Text, ...]]:
        """Get queues for this event broker.

        The preferred argument defining the RabbitMQ queues the `PikaEventBroker` should
        publish to is `queues` (as of Rasa Open Source version 1.8.2). This method
        can be removed in the future, and `self.queues` should just receive the value of
        the `queues` kwarg in the constructor.

        Args:
            queues_arg: Value of the supplied `queues` argument.

        Returns:
            Queues this event broker publishes to.

        Raises:
            `ValueError` if no valid `queues` argument was found.
        """
        if queues_arg and isinstance(queues_arg, (list, tuple)):
            return queues_arg

        if queues_arg and isinstance(queues_arg, str):
            logger.debug(
                f"Found a string value under the `queues` key of the Pika event broker "
                f"config. Please supply a list of queues under this key, even if it is "
                f"just a single one. See {DOCS_URL_PIKA_EVENT_BROKER}"
            )
            return [queues_arg]

        rasa.shared.utils.io.raise_warning(
            f"No `queues` argument provided. It is suggested to "
            f"explicitly specify a queue as described in "
            f"{DOCS_URL_PIKA_EVENT_BROKER}. "
            f"Using the default queue '{DEFAULT_QUEUE_NAME}' for now."
        )

        return [DEFAULT_QUEUE_NAME]

    @classmethod
    async def from_endpoint_config(
        cls,
        broker_config: Optional["EndpointConfig"],
        event_loop: Optional[AbstractEventLoop] = None,
    ) -> Optional["PikaEventBroker"]:
        """Creates broker. See the parent class for more information."""
        if broker_config is None:
            return None

        broker = cls(broker_config.url, **broker_config.kwargs, event_loop=event_loop)
        await broker.connect()

        return broker

    async def connect(self) -> None:
        """Connects to RabbitMQ."""
        self._connection = await self._connect()
        self._connection.reconnect_callbacks.add(self._publish_unpublished_messages)
        logger.info(f"RabbitMQ connection to '{self.host}' was established.")

        channel = await self._connection.channel()
        logger.debug(
            f"RabbitMQ channel was opened. "
            f"Declaring fanout exchange '{self.exchange_name}'."
        )

        self._exchange = await self._set_up_exchange(channel)

    def _configure_url(self) -> Optional[Text]:
        """Configures the URL to connect to RabbitMQ."""
        url = None

        if self.host.startswith("amqp"):
            parsed_host = urlparse(self.host)

            amqp_user = f"{self.username}:{self.password}"
            if amqp_user not in parsed_host.netloc:
                url = f"{parsed_host.scheme}://{amqp_user}@{parsed_host.netloc}"
            else:
                url = f"{parsed_host.scheme}://{parsed_host.netloc}"

            if str(self.port) not in url:
                url = f"{url}:{self.port}"

            if parsed_host.path:
                url = f"{url}{parsed_host.path}"

            if parsed_host.query:
                url = f"{url}?{parsed_host.query}"

        return url

    async def _connect(self) -> aio_pika.abc.AbstractRobustConnection:
        # The `url` parameter will take precedence over parameters like `login` or
        # `password`.
        url = self._configure_url()

        ssl_options = _create_rabbitmq_ssl_options(self.host)
        logger.info("Connecting to RabbitMQ ...")

        last_exception: Optional[Exception] = None
        for _ in range(self._connection_attempts):
            try:
                return await aio_pika.connect_robust(
                    url=url,
                    host=self.host,
                    port=self.port,
                    password=self.password,
                    login=self.username,
                    loop=self._loop,
                    ssl=ssl_options is not None,
                    ssl_options=ssl_options,
                )
            # All sorts of exception can happen until RabbitMQ is in a stable state
            except Exception as e:
                last_exception = e
                logger.debug(
                    f"Connecting to '{self.host}' failed with error '{e}'. "
                    f"Trying again."
                )
                await asyncio.sleep(self._retry_delay_in_seconds)

        last_exception = cast(Exception, last_exception)
        logger.error(
            f"Connecting to '{self.host}' failed with error '{last_exception}'."
        )
        raise last_exception

    def _publish_unpublished_messages(self, *_: Any, **__: Any) -> None:
        while self._unpublished_events:
            # Send unpublished messages
            message = self._unpublished_events.popleft()
            self.publish(message)
            logger.debug(
                f"Published message from queue of unpublished messages. "
                f"Remaining unpublished messages: {len(self._unpublished_events)}."
            )

    async def _set_up_exchange(
        self, channel: aio_pika.RobustChannel
    ) -> aio_pika.RobustExchange:
        exchange = await channel.declare_exchange(
            self.exchange_name, type=aio_pika.ExchangeType.FANOUT
        )

        await asyncio.gather(
            *[
                self._bind_queue(queue_name, channel, exchange)
                for queue_name in self.queues
            ]
        )

        return exchange

    @staticmethod
    async def _bind_queue(
        queue_name: Text, channel: aio_pika.RobustChannel, exchange: aio_pika.Exchange
    ) -> None:
        queue = await channel.declare_queue(queue_name, durable=True)

        await queue.bind(exchange, "")

    async def close(self) -> None:
        """Closes connection to RabbitMQ."""
        if not self._connection:
            return

        # Entering the context manager does nothing. Exiting closes the channels and
        # the connection.
        async with self._connection:
            logger.debug("Closing RabbitMQ connection.")

    def is_ready(self) -> bool:
        """Return `True` if a connection was established."""
        if self._connection is None:
            return False

        return not self._connection.is_closed

    def publish(
        self, event: Dict[Text, Any], headers: Optional[Dict[Text, Text]] = None
    ) -> None:
        """Publishes `event` to Pika queues.

        Args:
            event: Serialised event to be published.
            headers: Message headers to append to the published message. The headers
                can be retrieved in the consumer from the `headers` attribute of the
                message's `BasicProperties`.
        """
        # We need to save a reference to this background task to
        # make sure it doesn't disappear. See:
        # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
        task: asyncio.Task = self._loop.create_task(self._publish(event, headers))
        self._background_tasks.add(task)
        task.add_done_callback(self._background_tasks.discard)

    async def _publish(
        self, event: Dict[Text, Any], headers: Optional[Dict[Text, Text]] = None
    ) -> None:
        if self._exchange is None:
            return

        reduced_event = rasa.shared.core.events.remove_parse_data(event)

        try:
            await self._exchange.publish(self._message(event, headers), "")

            structlogger.debug(
                "pika.events.publish",
                event_info="Logging a reduced version of the Pika event",
                rabbitmq_exchange=RABBITMQ_EXCHANGE,
                host=self.host,
                rasa_event=reduced_event,
            )
        except Exception as e:
            structlogger.error(
                "pika.events.publish.failed",
                event_info=f"Failed to publish Pika event. Error: {e}",
                host=self.host,
            )
            if self.should_keep_unpublished_messages:
                self._unpublished_events.append(event)

            if self.raise_on_failure:
                self.close()  # type: ignore[unused-coroutine]
                raise e

    def _message(
        self, event: Dict[Text, Any], headers: Optional[Dict[Text, Text]]
    ) -> aio_pika.Message:
        body = json.dumps(event)
        return aio_pika.Message(
            bytes(body, DEFAULT_ENCODING),
            headers=headers,
            app_id=self.rasa_environment,
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
        )

    @cached_property
    def rasa_environment(self) -> Optional[Text]:
        """Get value of the `RASA_ENVIRONMENT` environment variable."""
        return os.environ.get("RASA_ENVIRONMENT")


def _create_rabbitmq_ssl_options(
    rabbitmq_host: Optional[Text] = None,
) -> Optional[Dict]:
    """Create RabbitMQ SSL options.

    Requires the following environment variables to be set:

        RABBITMQ_SSL_CLIENT_CERTIFICATE - path to the SSL client certificate (required)
        RABBITMQ_SSL_CLIENT_KEY - path to the SSL client key (required)

    Details on how to enable RabbitMQ TLS support can be found here:
    https://www.rabbitmq.com/ssl.html#enabling-tls

    Args:
        rabbitmq_host: RabbitMQ hostname.

    Returns:
        Optional SSL arguments for the RabbitMQ connection.
    """
    client_certificate_path = os.environ.get("RABBITMQ_SSL_CLIENT_CERTIFICATE")
    client_key_path = os.environ.get("RABBITMQ_SSL_CLIENT_KEY")

    if os.environ.get("RABBITMQ_SSL_CA_FILE"):
        rasa.shared.utils.io.raise_warning(
            "Specifying 'RABBITMQ_SSL_CA_FILE' via "
            "environment variables is no longer supported. Please specify this "
            "through the RabbitMQ URL parameter 'cacertfile' as described here: "
            "https://www.rabbitmq.com/uri-query-parameters.html "
        )

    if os.environ.get("RABBITMQ_SSL_KEY_PASSWORD"):
        rasa.shared.utils.io.raise_warning(
            "Specifying 'RABBITMQ_SSL_KEY_PASSWORD' via environment variables is no "
            "longer supported. Please use an unencrypted key file."
        )

    if client_certificate_path and client_key_path:
        logger.debug(f"Configuring SSL context for RabbitMQ host '{rabbitmq_host}'.")
        return {
            "certfile": client_certificate_path,
            "client_key_path": client_key_path,
            "cert_reqs": ssl.CERT_REQUIRED,
        }

    return None



from datetime import datetime
from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import ActionExecuted, SlotSet

from actions.database import Database


class ActionSessionStart(Action):
    def name(self) -> Text:
        return "action_session_start"

    def set_current_date(self) -> List[Dict[Text, Any]]:
        current_date = datetime.now().strftime("%d/%m/%Y")
        return [SlotSet("current_date", current_date)]

    def set_user_profile(self, username: str) -> List[Dict[Text, Any]]:
        db = Database()

        # Get user information
        user = db.get_user_by_name(username)

        if user:
            return [
                SlotSet("username", user["name"]),
                SlotSet("segment", user["segment"]),
                SlotSet("email_address", user["email"]),
                SlotSet("physical_address", user["address"]),
            ]
        else:
            return []

    def run(
        self, dispatcher, tracker: Tracker, domain: Dict[Text, Any]
    ) -> List[Dict[Text, Any]]:
        current_date_events = self.set_current_date()

        # Advanced option: It is possible to enable the chatbot with multiple
        # user profiles.
        # This option influences the interactions with the assistant.
        # The bot randomly assigns a user profile at the start of each session.
        # With this advance option: The demo bot will include three user profiles
        # with different payees:
        # John Smith: Payees: Robert (friend), James (son), Food Market (groceries)
        # Mary Brown: Payees: Richard (business partner), Susan (friend),
        # Electric Company (utilities)
        # Dan Young: Payees: Amy (colleague), Fitness Gym (gym), William (friend)
        #
        # The default profile is John Smith
        #
        # Each profile also has its own accounts—Current (Checking), Savings,
        # and Business
        #
        # Intructions: Just toggle the comments of the following 2 lines.
        username = "John Smith"
        # username = random.choice(["John Smith", "Mary Brown", "Dan Young"])

        user_profile_events = self.set_user_profile(username)

        events = (
            current_date_events
            + user_profile_events
            + [
                ActionExecuted("action_listen"),
            ]
        )

        return events



"""Pydantic models for request/response validation."""

from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, validator

from rasa.cli.scaffold import ProjectTemplateName
from rasa.shared.core.trackers import DialogueStateTracker


class PromptRequest(BaseModel):
    """Request model for prompt-to-bot endpoint."""

    prompt: str = Field(
        ..., min_length=1, max_length=10000, description="The skill description prompt"
    )
    client_id: Optional[str] = Field(
        None, max_length=255, description="Optional client identifier"
    )

    @validator("prompt")
    def validate_prompt(cls, v):
        if not v.strip():
            raise ValueError("Prompt cannot be empty or whitespace only")
        return v.strip()


class TemplateRequest(BaseModel):
    """Request model for template-to-bot endpoint."""

    template_name: ProjectTemplateName = Field(
        ...,
        description=(
            f"The template name to use ({ProjectTemplateName.supported_values()})"
        ),
    )
    client_id: Optional[str] = Field(
        None, max_length=255, description="Optional client identifier"
    )

    @validator("template_name")
    def validate_template_name(cls, v):
        if v not in ProjectTemplateName:
            raise ValueError(
                f"Template name must be one of {ProjectTemplateName.supported_values()}"
            )
        return v


class ChatMessage(BaseModel):
    """Model for chat messages."""

    type: str = Field(..., pattern="^(user|assistant)$")
    content: Union[str, List[Dict[str, Any]]] = Field(...)


class LLMBuilderRequest(BaseModel):
    """Request model for LLM builder endpoint."""

    messages: List[ChatMessage] = Field(..., min_items=1, max_items=50)


class LLMBuilderContext(BaseModel):
    """Context model for LLM builder endpoint."""

    tracker: Optional[DialogueStateTracker] = Field(None)
    bot_logs: str = Field("")
    chat_bot_files: Dict[str, str] = Field({})
    chat_history: List[ChatMessage] = Field([])

    class Config:
        """Config for LLMBuilderContext."""

        arbitrary_types_allowed = True


class BotDataUpdateRequest(BaseModel):
    """Request model for bot data updates."""

    domain_yml: Optional[str] = Field(None, alias="domain.yml")
    flows_yml: Optional[str] = Field(None, alias="flows.yml")
    config_yml: Optional[str] = Field(None, alias="config.yml")

    class Config:
        """Config for BotDataUpdateRequest."""

        allow_population_by_field_name = True


class ContentBlock(BaseModel):
    """Base model for content blocks."""

    type: str = Field(...)


class TextBlock(ContentBlock):
    """Text content block."""

    type: Literal["text"] = "text"
    text: str = Field(...)


class CodeBlock(ContentBlock):
    """Code content block."""

    type: Literal["code"] = "code"
    text: str = Field(...)
    language: Optional[str] = Field(None)


class FileBlock(ContentBlock):
    """File content block."""

    type: Literal["file"] = "file"
    file: str = Field(...)
    content: str = Field(...)


class LinkBlock(ContentBlock):
    """Link content block."""

    type: Literal["link"] = "link"
    text: str = Field(..., pattern=r"^https?://")


class LLMHelperResponse(BaseModel):
    """Response model for LLM helper."""

    content_blocks: List[Union[TextBlock, CodeBlock, FileBlock, LinkBlock]] = Field(...)


class ApiResponse(BaseModel):
    """Standard API response model."""

    status: str = Field(...)
    message: Optional[str] = Field(None)
    data: Optional[Dict[str, Any]] = Field(None)


class ApiErrorResponse(BaseModel):
    """API error response model."""

    status: Literal["error"] = "error"
    error: str = Field(...)
    details: Optional[Dict[str, Any]] = Field(None)


class ServerSentEvent(BaseModel):
    """Server-sent event model."""

    event: str = Field(...)
    data: Dict[str, Any] = Field(...)

    def format(self) -> str:
        """Format as SSE string."""
        import json

        return f"event: {self.event}\ndata: {json.dumps(self.data)}\n\n"


class ValidationResult(BaseModel):
    """Result of validation operation."""

    is_valid: bool = Field(...)
    errors: Optional[List[str]] = Field(None)
    warnings: Optional[List[str]] = Field(None)


class TrainingResult(BaseModel):
    """Result of training operation."""

    success: bool = Field(...)
    model_path: Optional[str] = Field(None)
    error: Optional[str] = Field(None)



"""Service for generating Rasa projects from prompts."""

import json
import os
import shutil
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, List, Optional

import structlog

from rasa.builder import config
from rasa.builder.exceptions import ProjectGenerationError, ValidationError
from rasa.builder.llm_service import get_skill_generation_messages, llm_service
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName, create_initial_project
from rasa.shared.core.flows import yaml_flows_io
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string
from rasa.utils.io import subpath

structlogger = structlog.get_logger()


class ProjectGenerator:
    """Service for generating Rasa projects from skill descriptions."""

    def __init__(self, project_folder: str):
        """Initialize the project generator with a folder for file persistence.

        Args:
            project_folder: Path to the folder where project files will be stored
        """
        self.project_folder = Path(project_folder)
        self.project_folder.mkdir(parents=True, exist_ok=True)

    def init_from_template(self, template: ProjectTemplateName):
        """Create the initial project files."""
        self.cleanup()
        create_initial_project(self.project_folder.as_posix(), template)

    async def generate_project_with_retries(
        self,
        skill_description: str,
        template: ProjectTemplateName,
        max_retries: Optional[int] = None,
    ) -> Dict[str, str]:
        """Generate a Rasa project with retry logic for validation failures.

        Args:
            skill_description: Natural language description of the skill
            rasa_config: Rasa configuration dictionary
            template: Project template to use for the initial project
            max_retries: Maximum number of retry attempts

        Returns:
            Dictionary of generated file contents (filename -> content)

        Raises:
            ProjectGenerationError: If generation fails after all retries
        """
        if max_retries is None:
            max_retries = config.MAX_RETRIES

        self.init_from_template(template)

        project_data = self._get_bot_data_for_llm()

        initial_messages = get_skill_generation_messages(
            skill_description, project_data
        )

        async def _generate_with_retry(
            messages: List[Dict[str, Any]], attempts_left: int
        ):
            try:
                # Generate project data using LLM
                project_data = await llm_service.generate_rasa_project(messages)

                # Update stored bot data
                self._update_bot_files_from_llm_response(project_data)

                bot_files = self.get_bot_files()
                structlogger.info(
                    "project_generator.generated_project",
                    attempts_left=attempts_left,
                    files=list(bot_files.keys()),
                )

                # Validate the generated project
                await self._validate_generated_project()

                structlogger.info(
                    "project_generator.validation_success", attempts_left=attempts_left
                )

                return bot_files

            except ValidationError as e:
                structlogger.error(
                    "project_generator.validation_error",
                    error=str(e),
                    attempts_left=attempts_left,
                )

                if attempts_left <= 0:
                    raise ProjectGenerationError(
                        f"Failed to generate valid Rasa project: {e}", max_retries
                    )

                # Create error feedback for next attempt
                error_feedback_messages = messages + [
                    {
                        "role": "assistant",
                        "content": json.dumps(project_data),
                    },
                    {
                        "role": "user",
                        "content": dedent(f"""
                            Previous attempt failed validation with error: {e}

                            Please fix the issues and generate a valid Rasa project.
                            Pay special attention to:
                            - Proper YAML syntax
                            - Required fields in domain and flows
                            - Consistent naming between flows and domain
                            - Valid slot types and mappings
                        """).strip(),
                    },
                ]

                return await _generate_with_retry(
                    error_feedback_messages, attempts_left - 1
                )

            except Exception as e:
                structlogger.error(
                    "project_generator.generation_error",
                    error=str(e),
                    attempts_left=attempts_left,
                )

                if attempts_left <= 0:
                    raise ProjectGenerationError(
                        f"Failed to generate Rasa project: {e}", max_retries
                    )

                # For non-validation errors, retry with original messages
                return await _generate_with_retry(initial_messages, attempts_left - 1)

        return await _generate_with_retry(initial_messages, max_retries)

    async def _validate_generated_project(self):
        """Validate the generated project using the validation service."""
        importer = self._create_importer()
        validation_error = await validate_project(importer)

        if validation_error:
            raise ValidationError(validation_error)

    def _create_importer(self) -> TrainingDataImporter:
        """Create a training data importer from the current bot files."""
        try:
            if (self.project_folder / "domain.yml").exists():
                domain_path = self.project_folder / "domain.yml"
            else:
                domain_path = self.project_folder / "domain"

            return TrainingDataImporter.load_from_config(
                config_path=self.project_folder / "config.yml",
                domain_path=domain_path,
                training_data_paths=[
                    self.project_folder / "data",
                ],
                args={},
            )

        except Exception as e:
            raise ValidationError(f"Failed to create importer: {e}")

    def get_bot_files(self) -> Dict[str, str]:
        """Get the current bot files by reading from disk."""
        bot_files = {}

        for file in self.project_folder.glob("**/*"):
            # Skip directories
            if not file.is_file():
                continue

            relative_path = file.relative_to(self.project_folder)

            # Skip hidden files and directories (any path component starting with '.')
            if any(part.startswith(".") for part in relative_path.parts):
                continue

            # exclude the project_folder / models folder
            if relative_path.parts[0] == "models":
                continue

            # Read file content and store with relative path as key
            bot_files[relative_path.as_posix()] = file.read_text(encoding="utf-8")

        return bot_files

    def _get_bot_data_for_llm(self) -> Dict[str, Any]:
        """Get the current bot data for the LLM."""
        file_importer = self._create_importer()

        # only include data created by the user (or the builder llm)
        # avoid including to many defaults that are not customized
        domain = file_importer.get_user_domain()
        flows = file_importer.get_user_flows()

        return {
            "domain": domain.as_dict(should_clean_json=True),
            "flows": yaml_flows_io.get_flows_as_json(flows, should_clean_json=True),
        }

    def _path_for_flow(self, flow_id: str) -> str:
        """Get the path for a flow."""
        if flow_id.startswith("pattern_"):
            return f"data/patterns/{flow_id}.yml"
        else:
            return f"data/flows/{flow_id}.yml"

    def _update_bot_files_from_llm_response(self, project_data: Dict[str, Any]):
        """Update the bot files with generated data by writing to disk."""
        files = {"domain.yml": dump_obj_as_yaml_to_string(project_data["domain"])}
        # split up flows into one file per flow in the /flows folder
        for flow_id, flow_data in project_data["flows"].get("flows", {}).items():
            flow_file_path = self._path_for_flow(flow_id)
            single_flow_file_data = {"flows": {flow_id: flow_data}}
            files[flow_file_path] = dump_obj_as_yaml_to_string(single_flow_file_data)

        # removes any other flows that the LLM didn't generate
        self._cleanup_flows()
        self.update_bot_files(files)

    def _cleanup_flows(self):
        """Cleanup the flows folder."""
        flows_folder = self.project_folder / "data" / "flows"
        if flows_folder.exists():
            shutil.rmtree(flows_folder)
        flows_folder.mkdir(parents=True, exist_ok=True)

    def update_bot_files(self, files: Dict[str, str]):
        """Update bot files with new content by writing to disk."""
        for filename, content in files.items():
            file_path = Path(subpath(self.project_folder, filename))
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(content, encoding="utf-8")

    def cleanup(self):
        """Cleanup the project folder."""
        # remove all the files and folders in the project folder resulting
        # in an empty folder
        for filename in os.listdir(self.project_folder):
            file_path = os.path.join(self.project_folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                structlogger.error(
                    "project_generator.cleanup_error",
                    error=str(e),
                    file_path=file_path,
                )



import asyncio
import copy
import functools
import importlib
import inspect
import logging
import os
import pkgutil
import sys
from types import ModuleType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Text,
    Type,
)

import rasa.shared.utils.io
from rasa.exceptions import MissingDependencyException
from rasa.shared.constants import DOCS_URL_MIGRATION_GUIDE
from rasa.shared.exceptions import ProviderClientValidationError, RasaException
from rasa.shared.utils.cli import print_success

logger = logging.getLogger(__name__)


def class_from_module_path(
    module_path: Text, lookup_path: Optional[Text] = None
) -> Type:
    """Given the module name and path of a class, tries to retrieve the class.

    The loaded class can be used to instantiate new objects.

    Args:
        module_path: either an absolute path to a Python class,
                     or the name of the class in the local / global scope.
        lookup_path: a path where to load the class from, if it cannot
                     be found in the local / global scope.

    Returns:
        a Python class

    Raises:
        ImportError, in case the Python class cannot be found.
        RasaException, in case the imported result is something other than a class
    """
    warn_and_exit_if_module_path_contains_rasa_plus(module_path, lookup_path)

    klass = None
    if "." in module_path:
        module_name, _, class_name = module_path.rpartition(".")
        m = importlib.import_module(module_name)
        klass = getattr(m, class_name, None)
    elif lookup_path:
        # try to import the class from the lookup path
        m = importlib.import_module(lookup_path)
        klass = getattr(m, module_path, None)

    if klass is None:
        raise ImportError(f"Cannot retrieve class from path {module_path}.")

    if not inspect.isclass(klass):
        raise RasaException(
            f"`class_from_module_path()` is expected to return a class, "
            f"but for {module_path} we got a {type(klass)}."
        )
    return klass


def import_package_modules(package_name: str) -> List[ModuleType]:
    """Import all modules in a package."""
    package = importlib.import_module(package_name)
    return [
        importlib.import_module(f"{package_name}.{module_name}")
        for _, module_name, _ in pkgutil.iter_modules(package.__path__)
    ]


def all_subclasses(cls: Any) -> List[Any]:
    """Returns all known (imported) subclasses of a class."""
    classes = cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in all_subclasses(s)
    ]

    return [subclass for subclass in classes if not inspect.isabstract(subclass)]


def module_path_from_instance(inst: Any) -> Text:
    """Return the module path of an instance's class."""
    return inst.__module__ + "." + inst.__class__.__name__


def sort_list_of_dicts_by_first_key(dicts: List[Dict]) -> List[Dict]:
    """Sorts a list of dictionaries by their first key."""
    return sorted(dicts, key=lambda d: next(iter(d.keys())))


def cached_method(f: Callable[..., Any]) -> Callable[..., Any]:
    """Caches method calls based on the call's `args` and `kwargs`.
    Works for `async` and `sync` methods. Don't apply this to functions.
    Args:
        f: The decorated method whose return value should be cached.
    Returns:
        The return value which the method gives for the first call with the given
        arguments.
    """
    assert "self" in arguments_of(f), "This decorator can only be used with methods."

    class Cache:
        """Helper class to abstract the caching details."""

        def __init__(self, caching_object: object, args: Any, kwargs: Any) -> None:
            self.caching_object = caching_object
            self.cache = getattr(caching_object, self._cache_name(), {})
            # noinspection PyUnresolvedReferences
            self.cache_key = functools._make_key(args, kwargs, typed=False)

        def _cache_name(self) -> Text:
            return f"_cached_{self.caching_object.__class__.__name__}_{f.__name__}"

        def is_cached(self) -> bool:
            return self.cache_key in self.cache

        def cache_result(self, result: Any) -> None:
            self.cache[self.cache_key] = result
            setattr(self.caching_object, self._cache_name(), self.cache)

        def cached_result(self) -> Any:
            return self.cache[self.cache_key]

    if asyncio.iscoroutinefunction(f):

        @functools.wraps(f)
        async def decorated(self: object, *args: Any, **kwargs: Any) -> Any:
            cache = Cache(self, args, kwargs)
            if not cache.is_cached():
                # Store the task immediately so that other concurrent calls of the
                # method can re-use the same task and don't schedule a second execution.
                to_cache = asyncio.ensure_future(f(self, *args, **kwargs))
                cache.cache_result(to_cache)
            return await cache.cached_result()

        return decorated
    else:

        @functools.wraps(f)
        def decorated(self: object, *args: Any, **kwargs: Any) -> Any:
            cache = Cache(self, args, kwargs)
            if not cache.is_cached():
                to_cache = f(self, *args, **kwargs)
                cache.cache_result(to_cache)
            return cache.cached_result()

        return decorated


def transform_collection_to_sentence(collection: Collection[Text]) -> Text:
    """Transforms e.g. a list like ['A', 'B', 'C'] into a sentence 'A, B and C'."""
    x = list(collection)
    if len(x) >= 2:
        return ", ".join(map(str, x[:-1])) + " and " + x[-1]
    return "".join(collection)


def minimal_kwargs(
    kwargs: Dict[Text, Any], func: Callable, excluded_keys: Optional[List] = None
) -> Dict[Text, Any]:
    """Returns only the kwargs which are required by a function.

    Keys, contained in the exception list, are not included.

    Args:
        kwargs: All available kwargs.
        func: The function which should be called.
        excluded_keys: Keys to exclude from the result.

    Returns:
        Subset of kwargs which are accepted by `func`.

    """
    excluded_keys = excluded_keys or []

    possible_arguments = arguments_of(func)

    return {
        k: v
        for k, v in kwargs.items()
        if k in possible_arguments and k not in excluded_keys
    }


def mark_as_experimental_feature(feature_name: Text) -> None:
    """Warns users that they are using an experimental feature."""
    logger.warning(
        f"The {feature_name} is currently experimental and might change or be "
        "removed in the future 🔬 Please share your feedback on it in the "
        "forum (https://forum.rasa.com) to help us make this feature "
        "ready for production."
    )


def mark_as_beta_feature(feature_name: Text) -> None:
    """Warns users that they are using a beta feature."""
    logger.warning(
        f"🔬 Beta Feature: {feature_name} is in beta. It may have unexpected "
        "behaviour and might be changed in the future."
    )


def arguments_of(func: Callable) -> List[Text]:
    """Return the parameters of the function `func` as a list of names."""
    import inspect

    return list(inspect.signature(func).parameters.keys())


def extract_duplicates(list1: List[Any], list2: List[Any]) -> List[Any]:
    """Extracts duplicates from two lists."""
    if list1:
        dict1 = {
            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in list1
        }
    else:
        dict1 = {}

    if list2:
        dict2 = {
            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in list2
        }
    else:
        dict2 = {}

    set1 = set(dict1.keys())
    set2 = set(dict2.keys())
    dupes = set1.intersection(set2)
    return sorted(list(dupes))


def clean_duplicates(dupes: Dict[Text, Any]) -> Dict[Text, Any]:
    """Removes keys for empty values."""
    duplicates = dupes.copy()
    for k in dupes:
        if not dupes[k]:
            duplicates.pop(k)

    return duplicates


def merge_dicts(
    tempDict1: Dict[Text, Any],
    tempDict2: Dict[Text, Any],
    override_existing_values: bool = False,
) -> Dict[Text, Any]:
    """Merges two dicts."""
    if override_existing_values:
        merged_dicts, b = tempDict1.copy(), tempDict2.copy()

    else:
        merged_dicts, b = tempDict2.copy(), tempDict1.copy()
    merged_dicts.update(b)
    return merged_dicts


def merge_lists(
    list1: List[Any], list2: List[Any], override: bool = False
) -> List[Any]:
    """Merges two lists."""
    return sorted(list(set(list1 + list2)))


def merge_lists_of_dicts(
    dict_list1: List[Dict],
    dict_list2: List[Dict],
    override_existing_values: bool = False,
) -> List[Dict]:
    """Merges two dict lists."""
    dict1 = {
        (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in dict_list1
    }
    dict2 = {
        (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in dict_list2
    }
    merged_dicts = merge_dicts(dict1, dict2, override_existing_values)
    return list(merged_dicts.values())


def partial_merge_list(
    self_list: List[Any],
    other_list: List[Any],
    is_same_item_fn: Callable[[Any, Any], bool],
) -> List[Any]:
    """Merges two lists based on a custom intersection logic."""
    matched_other_indices = set()
    result = []

    for s_item in self_list:
        match = next(
            (
                (i, o_item)
                for i, o_item in enumerate(other_list)
                if i not in matched_other_indices and is_same_item_fn(s_item, o_item)
            ),
            None,
        )
        if match:
            i, o_item = match
            result.append(copy.deepcopy(o_item))
            matched_other_indices.add(i)
        else:
            result.append(copy.deepcopy(s_item))

    return result


def partial_merge_dict(
    self_d: Dict[Text, Any], other_d: Dict[Text, Any]
) -> Dict[Text, Any]:
    """Merges two dictionaries based on a custom intersection logic."""
    merged = {}
    for k, v in self_d.items():
        if k in other_d:
            merged[k] = copy.deepcopy(other_d[k])
        else:
            merged[k] = copy.deepcopy(v)
    return merged


def warn_and_exit_if_module_path_contains_rasa_plus(
    module_path: Text, lookup_path: Optional[str] = None
) -> None:
    """Warns and exits if the module path contains `rasa_plus`.

    Args:
        module_path: The module path to check.
        lookup_path: The lookup path to check.
    """
    if "rasa_plus" in module_path.lower() or (
        lookup_path and "rasa_plus" in lookup_path.lower()
    ):
        rasa.shared.utils.io.raise_warning(
            f"Your module path '{module_path}' contains 'rasa_plus'. "
            f"The path to this Rasa Pro component has changed, please "
            f"follow the migration guide in the official documentation "
            f"to update your code: ",
            docs=DOCS_URL_MIGRATION_GUIDE,
        )
        sys.exit(1)


def validate_environment(
    required_env_vars: Sequence[str],
    required_packages: Sequence[str],
    component_name: str,
) -> None:
    """Make sure all needed requirements for a component are met.
    Args:
         required_env_vars: List of environment variables that should be set
         required_packages: List of packages that should be installed
         component_name: component name that needs the requirements
    """
    for e in required_env_vars:
        if not os.environ.get(e):
            raise ProviderClientValidationError(
                f"Missing environment variable for {component_name}: {e}"
            )
    for p in required_packages:
        try:
            importlib.import_module(p)
        except ImportError:
            raise MissingDependencyException(
                f"Missing package for {component_name}: {p}"
            )


def display_research_study_prompt() -> None:
    """Logs a prompt to get users to participate in a UX Improvement research study."""
    separator = "-" * 54
    message = f"""
{separator}
Want to help shape the future of Rasa Pro?
Share your feedback in a short conversation with our team.
Sign up at: https://feedback.rasa.com
{separator}
"""
    print_success(message)



import argparse
import importlib
import json
import os
import sys
import time
from pathlib import Path
from types import FrameType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, Union, overload

import randomname
import structlog

import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa import telemetry
from rasa.exceptions import ModelNotFound
from rasa.shared.constants import (
    ASSISTANT_ID_DEFAULT_VALUE,
    ASSISTANT_ID_KEY,
    DEFAULT_CONFIG_PATH,
)
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.common import display_research_study_prompt
from rasa.shared.utils.yaml import read_config_file
from rasa.utils.io import write_yaml

if TYPE_CHECKING:
    from questionary import Question
    from typing_extensions import Literal

    from rasa.validator import Validator

structlogger = structlog.get_logger()

FREE_TEXT_INPUT_PROMPT = "Type out your own message..."


@overload
def get_validated_path(
    current: Optional[Union[Path, Text]],
    parameter: Text,
    default: Optional[Union[Path, Text, List[Text]]] = ...,
    none_is_valid: "Literal[False]" = ...,
) -> Union[Path, Text]: ...


@overload
def get_validated_path(
    current: Optional[Union[Path, Text]],
    parameter: Text,
    default: Optional[Union[Path, Text, List[Text]]] = ...,
    none_is_valid: "Literal[True]" = ...,
) -> Optional[Union[Path, Text]]: ...


def get_validated_path(
    current: Optional[Union[Path, Text]],
    parameter: Text,
    default: Optional[Union[Path, Text, List[Text]]] = None,
    none_is_valid: bool = False,
) -> Optional[Union[Path, Text]]:
    """Checks whether a file path or its default value is valid and returns it.

    Args:
        current: The parsed value.
        parameter: The name of the parameter.
        default: one or multiple default values of the parameter.
        none_is_valid: `True` if `None` is valid value for the path,
                        else `False``

    Returns:
        The current value if valid,
        otherwise one of the default values of the argument if valid,
        otherwise `None` if allowed,
        otherwise raises an error and exits.
    """
    if current and os.path.exists(current):
        return current

    if parameter == "model":
        raise ModelNotFound(
            f"The provided model path '{current}' could not be found. "
            "Provide an existing model path."
        )

    # try to find a valid option among the defaults
    if isinstance(default, str) or isinstance(default, Path):
        default_options = [str(default)]
    elif isinstance(default, list):
        default_options = default
    else:
        default_options = []

    valid_options = (option for option in default_options if os.path.exists(option))
    chosen_option = next(valid_options, None)

    # warn and log if user-chosen parameter wasn't found and thus overwritten
    if chosen_option:
        shared_info = f"Using default location '{chosen_option}' instead."
        if current is None:
            structlogger.debug(
                "cli.get_validated_path.parameter_not_set",
                parameter=parameter,
                event_info=(f"Parameter '{parameter}' was not set. {shared_info}"),
            )
        elif current not in default_options:
            structlogger.warn(
                "cli.get_validated_path.path_does_not_exists",
                path=current,
                event_info=(
                    f"The path '{current}' does not seem to exist. {shared_info}"
                ),
            )

    if chosen_option is None and not none_is_valid:
        cancel_cause_not_found(current, parameter, default)

    return chosen_option


def missing_config_keys(
    path: Union["Path", Text], mandatory_keys: List[Text]
) -> List[Text]:
    """Checks whether the config file at `path` contains the `mandatory_keys`.

    Args:
        path: The path to the config file.
        mandatory_keys: A list of mandatory config keys.

    Returns:
        The list of missing config keys.
    """
    if not os.path.exists(path):
        return mandatory_keys

    config_data = read_config_file(path)

    return [k for k in mandatory_keys if k not in config_data or config_data[k] is None]


def validate_assistant_id_in_config(config_file: Union["Path", Text]) -> None:
    """Verifies that the assistant_id key exists and has a unique value in config.

    Issues a warning if the key does not exist or has the default value and replaces it
    with a pseudo-random string value.
    """
    config_data = read_config_file(config_file, reader_type=["safe", "rt"])
    assistant_id = config_data.get(ASSISTANT_ID_KEY)

    if assistant_id is None or assistant_id == ASSISTANT_ID_DEFAULT_VALUE:
        structlogger.warn(
            "cli.validate_assistant_id_in_config.missing_unique_assistant_id_key",
            config=config_file,
            missing_key=ASSISTANT_ID_KEY,
            event_info=(
                f"The config file '{config_file!s}' is "
                f"missing a unique value for the "
                f"'{ASSISTANT_ID_KEY}' mandatory key. "
                f"Proceeding with generating a random "
                f"value and overwriting the '{ASSISTANT_ID_KEY}'"
                f" in the config file."
            ),
        )

        # add random value for assistant id, overwrite config file
        time_format = "%Y%m%d-%H%M%S"
        config_data[ASSISTANT_ID_KEY] = (
            f"{time.strftime(time_format)}-{randomname.get_name()}"
        )

        write_yaml(data=config_data, target=config_file, should_preserve_key_order=True)

    return


def validate_config_path(
    config: Optional[Union[Text, "Path"]],
    default_config: Text = DEFAULT_CONFIG_PATH,
) -> Text:
    """Verifies that the config path exists.

    Exit if the config file does not exist.

    Args:
        config: Path to the config file.
        default_config: default config to use if the file at `config` doesn't exist.

    Returns: The path to the config file.
    """
    config = rasa.cli.utils.get_validated_path(config, "config", default_config)

    if not config or not os.path.exists(config):
        structlogger.error(
            "cli.validate_config_path.does_not_exists",
            config=config,
            event_info=(
                f"The config file '{config}' does not exist. "
                f"Use '--config' to specify a valid config file."
            ),
        )
        display_research_study_prompt()
        sys.exit(1)

    return str(config)


def validate_mandatory_config_keys(
    config: Union[Text, "Path"],
    mandatory_keys: List[Text],
) -> Text:
    """Get a config from a config file and check if it is valid.

    Exit if the config isn't valid.

    Args:
        config: Path to the config file.
        mandatory_keys: The keys that have to be specified in the config file.

    Returns: The path to the config file if the config is valid.
    """
    missing_keys = set(rasa.cli.utils.missing_config_keys(config, mandatory_keys))
    if missing_keys:
        structlogger.error(
            "cli.validate_mandatory_config_keys.missing_keys",
            config=config,
            missing_keys=missing_keys,
            event_info=(
                "The config file '{}' is missing mandatory parameters: "
                "'{}'. Add missing parameters to config file and try again.".format(
                    config, "', '".join(missing_keys)
                )
            ),
        )
        display_research_study_prompt()
        sys.exit(1)

    return str(config)


def get_validated_config(
    config: Optional[Union[Text, "Path"]],
    mandatory_keys: List[Text],
    default_config: Text = DEFAULT_CONFIG_PATH,
) -> Text:
    """Validates config and returns path to validated config file."""
    config = validate_config_path(config, default_config)
    validate_assistant_id_in_config(config)

    config = validate_mandatory_config_keys(config, mandatory_keys)

    return config


def validate_files(
    fail_on_warnings: bool,
    max_history: Optional[int],
    importer: TrainingDataImporter,
    stories_only: bool = False,
    flows_only: bool = False,
    translations_only: bool = False,
) -> None:
    """Validates either the story structure or the entire project.

    Args:
        fail_on_warnings: `True` if the process should exit with a non-zero status
        max_history: The max history to use when validating the story structure.
        importer: The `TrainingDataImporter` to use to load the training data.
        stories_only: If `True`, only the story structure is validated.
        flows_only: If `True`, only the flows are validated.
        translations_only: If `True`, only the translations data is validated.
    """
    from rasa.validator import Validator

    validator = Validator.from_importer(importer)

    if stories_only:
        all_good = _validate_story_structure(validator, max_history, fail_on_warnings)
    elif flows_only:
        all_good = validator.verify_flows()
    elif translations_only:
        all_good = validator.verify_translations()
    else:
        if importer.get_domain().is_empty():
            structlogger.error(
                "cli.validate_files.empty_domain",
                event_info="Encountered empty domain during validation.",
            )
            display_research_study_prompt()
            sys.exit(1)

        valid_domain = _validate_domain(validator)
        valid_nlu = _validate_nlu(validator, fail_on_warnings)
        valid_stories = _validate_story_structure(
            validator, max_history, fail_on_warnings
        )
        valid_flows = validator.verify_flows()
        valid_translations = validator.verify_translations(summary_mode=True)
        valid_CALM_slot_mappings = validator.validate_CALM_slot_mappings()

        all_good = (
            valid_domain
            and valid_nlu
            and valid_stories
            and valid_flows
            and valid_translations
            and valid_CALM_slot_mappings
        )

    validator.warn_if_config_mandatory_keys_are_not_set()

    telemetry.track_validate_files(all_good)
    if not all_good:
        structlogger.error(
            "cli.validate_files.project_validation_error",
            event_info="Project validation completed with errors.",
        )
        display_research_study_prompt()
        sys.exit(1)


def _validate_domain(validator: "Validator") -> bool:
    valid_domain_validity = validator.verify_domain_validity()
    valid_actions_in_stories_rules = validator.verify_actions_in_stories_rules()
    valid_forms_in_stories_rules = validator.verify_forms_in_stories_rules()
    valid_form_slots = validator.verify_form_slots()
    valid_slot_mappings = validator.verify_slot_mappings()
    valid_responses = validator.check_for_no_empty_parenthesis_in_responses()
    valid_buttons = validator.validate_button_payloads()
    valid_slot_validation = validator.verify_slot_validation()
    valid_conditional_responses = (
        validator.validate_conditional_response_variation_predicates()
    )
    return (
        valid_domain_validity
        and valid_actions_in_stories_rules
        and valid_forms_in_stories_rules
        and valid_form_slots
        and valid_slot_mappings
        and valid_responses
        and valid_buttons
        and valid_slot_validation
        and valid_conditional_responses
    )


def _validate_nlu(validator: "Validator", fail_on_warnings: bool) -> bool:
    return validator.verify_nlu(not fail_on_warnings)


def _validate_story_structure(
    validator: "Validator", max_history: Optional[int], fail_on_warnings: bool
) -> bool:
    # Check if a valid setting for `max_history` was given
    if isinstance(max_history, int) and max_history < 1:
        raise argparse.ArgumentTypeError(
            f"The value of `--max-history {max_history}` " f"is not a positive integer."
        )

    return validator.verify_story_structure(
        not fail_on_warnings, max_history=max_history
    )


def cancel_cause_not_found(
    current: Optional[Union["Path", Text]],
    parameter: Text,
    default: Optional[Union["Path", Text, List[Text]]],
) -> None:
    """Exits with an error because the given path was not valid.

    Args:
        current: The path given by the user.
        parameter: The name of the parameter.
        default: The default value of the parameter.

    """
    default_clause = ""
    if default and isinstance(default, str):
        default_clause = f"use the default location ('{default}') or"
    elif default and isinstance(default, list):
        default_clause = f"use one of the default locations ({', '.join(default)}) or"

    structlogger.error(
        "cli.path_does_not_exist",
        path=current,
        event_info=(
            f"The path '{current}' does not exist. "
            f"Please make sure to {default_clause} specify it "
            f"with '--{parameter}'."
        ),
    )
    display_research_study_prompt()
    sys.exit(1)


def parse_last_positional_argument_as_model_path() -> None:
    """Fixes the parsing of a potential positional model path argument."""
    if (
        len(sys.argv) >= 2
        # support relevant commands ...
        and sys.argv[1] in ["run", "shell", "interactive"]
        # but avoid interpreting subparser commands as model paths
        and sys.argv[1:] != ["run", "actions"]
        and not sys.argv[-2].startswith("-")
        and os.path.exists(sys.argv[-1])
    ):
        sys.argv.append(sys.argv[-1])
        sys.argv[-2] = "--model"


def button_to_string(button: Dict[Text, Any], idx: int = 0) -> Text:
    """Create a string representation of a button."""
    title = button.pop("title", "")

    if "payload" in button:
        payload = " ({})".format(button.pop("payload"))
    else:
        payload = ""

    # if there are any additional attributes, we append them to the output
    if button:
        details = " - {}".format(json.dumps(button, sort_keys=True))
    else:
        details = ""

    button_string = "{idx}: {title}{payload}{details}".format(
        idx=idx + 1, title=title, payload=payload, details=details
    )

    return button_string


def element_to_string(element: Dict[Text, Any], idx: int = 0) -> Text:
    """Create a string representation of an element."""
    title = element.pop("title", "")

    element_string = "{idx}: {title} - {element}".format(
        idx=idx + 1, title=title, element=json.dumps(element, sort_keys=True)
    )

    return element_string


def button_choices_from_message_data(
    message: Dict[Text, Any], allow_free_text_input: bool = True
) -> List[Text]:
    """Return list of choices to present to the user.

    If allow_free_text_input is True, an additional option is added
    at the end along with the response buttons that allows the user
    to type in free text.
    """
    choices = [
        button_to_string(button, idx)
        for idx, button in enumerate(message.get("buttons"))
    ]
    if allow_free_text_input:
        choices.append(FREE_TEXT_INPUT_PROMPT)
    return choices


async def payload_from_button_question(button_question: "Question") -> Text:
    """Prompt user with a button question and returns the nlu payload."""
    response = await button_question.ask_async()
    if response != FREE_TEXT_INPUT_PROMPT:
        # Extract intent slash command if it's a button
        response = response[response.rfind("(") + 1 : response.rfind(")")]
    return response


def signal_handler(_: int, __: FrameType) -> None:
    """Kills Rasa when OS signal is received."""
    print("Goodbye 👋")
    sys.exit(0)


def warn_if_rasa_plus_package_installed() -> None:
    """Issue a user warning in case the `rasa_plus` package is installed."""
    rasa_plus_package = "rasa_plus"
    if importlib.util.find_spec(rasa_plus_package) is not None:
        rasa.shared.utils.io.raise_warning(
            f"{rasa_plus_package} python package is no longer necessary "
            f"for using Rasa Pro. Please uninstall it.",
            UserWarning,
        )


def check_if_studio_command() -> bool:
    """Checks if the command is a Rasa Studio command."""
    return len(sys.argv) >= 2 and sys.argv[1] == "studio"


def get_e2e_results_file_name(
    results_output_path: Path,
    result_type: str,
) -> str:
    """Returns the name of the e2e results file."""
    if results_output_path.is_dir():
        file_name = results_output_path / f"e2e_results_{result_type}.yml"
    else:
        parent = results_output_path.parent
        stem = results_output_path.stem
        file_name = parent / f"{stem}_{result_type}.yml"

    return str(file_name)


def is_skip_validation_flag_set() -> bool:
    """Checks if the skip validation flag is set."""
    return "--skip-validation" in sys.argv



import asyncio
import audioop
import base64
import json
import uuid
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Text,
    Tuple,
)

import structlog

from rasa.core.channels import UserMessage
from rasa.core.channels.socketio import SocketBlueprint, SocketIOInput
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import call_state
from rasa.core.channels.voice_stream.tts import TTSEngine
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)
from rasa.hooks import hookimpl
from rasa.plugin import plugin_manager
from rasa.shared.core.constants import ACTION_LISTEN_NAME
from rasa.shared.core.events import ActionExecuted
from rasa.shared.core.trackers import EventVerbosity

if TYPE_CHECKING:
    from sanic import Sanic, Websocket  # type: ignore[attr-defined]
    from socketio import AsyncServer

    from rasa.core.channels.channel import InputChannel, UserMessage
    from rasa.shared.core.trackers import DialogueStateTracker


structlogger = structlog.get_logger()


def tracker_as_dump(tracker: "DialogueStateTracker") -> Dict[str, Any]:
    """Create a dump of the tracker state."""
    from rasa.shared.core.trackers import get_trackers_for_conversation_sessions

    multiple_tracker_sessions = get_trackers_for_conversation_sessions(tracker)

    if 0 <= len(multiple_tracker_sessions) <= 1:
        last_tracker = tracker
    else:
        last_tracker = multiple_tracker_sessions[-1]

    # TODO: this is a bug: the bridge converts this back to json, but it
    # should be json in the first place
    return last_tracker.current_state(EventVerbosity.AFTER_RESTART)


def does_need_action_prediction(tracker: "DialogueStateTracker") -> bool:
    """Check if the tracker needs an action prediction."""
    return (
        len(tracker.events) == 0
        or not isinstance(tracker.events[-1], ActionExecuted)
        or tracker.events[-1].action_name != ACTION_LISTEN_NAME
    )


class StudioTrackerUpdatePlugin:
    """Plugin for publishing tracker updates a socketio channel."""

    def __init__(self, socket_channel: "StudioChatInput") -> None:
        self.socket_channel = socket_channel
        self.tasks: List[asyncio.Task] = []

    def _cancel_tasks(self) -> None:
        """Cancel all remaining tasks."""
        for task in self.tasks:
            task.cancel()
        self.tasks = []

    def _cleanup_tasks(self) -> None:
        """Remove tasks that have already completed."""
        self.tasks = [task for task in self.tasks if not task.done()]

    @hookimpl  # type: ignore[misc]
    def after_new_user_message(self, tracker: "DialogueStateTracker") -> None:
        """Triggers a tracker update notification after a new user message."""
        self.handle_tracker_update(tracker)

    @hookimpl  # type: ignore[misc]
    def after_action_executed(self, tracker: "DialogueStateTracker") -> None:
        """Triggers a tracker update notification after an action is executed."""
        self.handle_tracker_update(tracker)

    def handle_tracker_update(self, tracker: "DialogueStateTracker") -> None:
        """Handles a tracker update when triggered by a hook."""
        structlogger.info(
            "studio_chat.after_tracker_update", sender_id=tracker.sender_id
        )
        # directly create a dump to avoid the tracker getting modified by another
        # function before it gets published (since the publishing is scheduled
        # as an async task)
        tracker_dump = tracker_as_dump(tracker)
        task = asyncio.create_task(
            self.socket_channel.publish_tracker_update(tracker.sender_id, tracker_dump)
        )
        self.tasks.append(task)
        self._cleanup_tasks()

    @hookimpl  # type: ignore[misc]
    def after_server_stop(self) -> None:
        """Cancels all remaining tasks when the server stops."""
        self._cancel_tasks()


class StudioChatInput(SocketIOInput, VoiceInputChannel):
    """Input channel for the communication between Rasa Studio and Rasa Pro."""

    requires_voice_license = False

    @classmethod
    def name(cls) -> Text:
        return "studio_chat"

    def __init__(
        self,
        server_url: str,
        asr_config: Dict,
        tts_config: Dict,
        user_message_evt: Text = "user_uttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        session_persistence: bool = False,
        socketio_path: Optional[Text] = "/socket.io",
        jwt_key: Optional[Text] = None,
        jwt_method: Optional[Text] = "HS256",
        metadata_key: Optional[Text] = "metadata",
    ) -> None:
        """Creates a `StudioChatInput` object."""
        from rasa.core.agent import Agent

        self.agent: Optional[Agent] = None
        self.latest_tracker_session_id = None

        # Initialize the SocketIO input channel
        SocketIOInput.__init__(
            self,
            user_message_evt=user_message_evt,
            bot_message_evt=bot_message_evt,
            namespace=namespace,
            session_persistence=session_persistence,
            socketio_path=socketio_path,
            jwt_key=jwt_key,
            jwt_method=jwt_method,
            metadata_key=metadata_key,
        )

        # Initialize the Voice Input Channel
        VoiceInputChannel.__init__(
            self,
            server_url=server_url,
            asr_config=asr_config,
            tts_config=tts_config,
        )

        # Dictionaries to manage active connections and background tasks
        # `active_connections` holds the active voice sessions
        # `background_tasks` holds the asyncio tasks for voice streaming
        self.active_connections: Dict[str, SocketIOVoiceWebsocketAdapter] = {}
        self.background_tasks: Dict[str, asyncio.Task] = {}

        self._register_tracker_update_hook()

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> "InputChannel":
        """Creates a StudioChatInput channel from credentials."""
        credentials = credentials or {}

        return cls(
            # Voice specific parameters
            server_url=credentials.get("server_url", ""),
            asr_config=credentials.get("asr", {}),
            tts_config=credentials.get("tts", {}),
            # SocketIO parameters
            user_message_evt=credentials.get("user_message_evt", "user_uttered"),
            bot_message_evt=credentials.get("bot_message_evt", "bot_uttered"),
            namespace=credentials.get("namespace"),
            session_persistence=credentials.get("session_persistence", False),
            socketio_path=credentials.get("socketio_path", "/socket.io"),
            jwt_key=credentials.get("jwt_key"),
            jwt_method=credentials.get("jwt_method", "HS256"),
            metadata_key=credentials.get("metadata_key", "metadata"),
        )

    def _register_tracker_update_hook(self) -> None:
        plugin_manager().register(StudioTrackerUpdatePlugin(self))

    async def on_tracker_updated(self, tracker: "DialogueStateTracker") -> None:
        """Triggers a tracker update notification after a change to the tracker."""
        await self.publish_tracker_update(tracker.sender_id, tracker_as_dump(tracker))

    async def publish_tracker_update(self, sender_id: str, tracker_dump: Dict) -> None:
        """Publishes a tracker update notification to the websocket."""
        if not self.sio:
            structlogger.error("studio_chat.on_tracker_updated.sio_not_initialized")
            return

        # we need the latest session id to use it for the llm helper to get the
        # most recent conversation the user had with the bot.
        self.latest_tracker_session_id = sender_id

        await self.sio.emit("tracker", tracker_dump, room=sender_id)

    async def on_message_proxy(
        self,
        on_new_message: Callable[["UserMessage"], Awaitable[Any]],
        message: "UserMessage",
    ) -> None:
        """Proxies the on_new_message call to the underlying channel.

        Triggers a tracker update notification after processing the message.
        """
        try:
            await on_new_message(message)
        except Exception as e:
            structlogger.exception(
                "studio_chat.on_new_message.error",
                error=str(e),
                sender_id=message.sender_id,
            )

        if not self.agent:
            structlogger.error("studio_chat.on_message_proxy.agent_not_initialized")
            return

        tracker = await self.agent.tracker_store.retrieve(message.sender_id)
        if tracker is None:
            structlogger.error("studio_chat.on_message_proxy.tracker_not_found")
            return

        await self.on_tracker_updated(tracker)

    async def handle_tracker_update(self, sid: str, data: Dict) -> None:
        from rasa.shared.core.trackers import DialogueStateTracker

        structlogger.debug(
            "studio_chat.sio.handle_tracker_update",
            sid=sid,
            sender_id=data["sender_id"],
        )
        if self.agent is None:
            structlogger.error("studio_chat.sio.agent_not_initialized")
            return None

        if not (domain := self.agent.domain):
            structlogger.error("studio_chat.sio.domain_not_initialized")
            return None

        async with self.agent.lock_store.lock(data["sender_id"]):
            tracker = DialogueStateTracker.from_dict(
                data["sender_id"], data["events"], domain.slots
            )

            # will override an existing tracker with the same id!
            await self.agent.tracker_store.save(tracker)

            processor = self.agent.processor
            if processor and does_need_action_prediction(tracker):
                output_channel = self.get_output_channel()

                await processor._run_prediction_loop(output_channel, tracker)
                await self.agent.tracker_store.save(tracker)

        await self.on_tracker_updated(tracker)

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        """Voice method to convert channel bytes to RasaAudioBytes."""
        return RasaAudioBytes(audioop.lin2ulaw(input_bytes, 4))

    async def collect_call_parameters(
        self, channel_websocket: "Websocket"
    ) -> Optional[CallParameters]:
        """Voice method to collect call parameters."""
        session_id = channel_websocket.session_id
        return CallParameters(session_id, "local", "local", stream_id=session_id)

    def map_input_message(
        self,
        message: Any,
        ws: "Websocket",
    ) -> VoiceChannelAction:
        """Voice method to map websocket messages to actions."""
        if "audio" in message:
            channel_bytes = base64.b64decode(message["audio"])
            audio_bytes = self.channel_bytes_to_rasa_audio_bytes(channel_bytes)
            return NewAudioAction(audio_bytes)
        elif "marker" in message:
            if message["marker"] == call_state.latest_bot_audio_id:
                # Just finished streaming last audio bytes
                call_state.is_bot_speaking = False  # type: ignore[attr-defined]
                if call_state.should_hangup:
                    structlogger.debug(
                        "studio_chat.hangup", marker=call_state.latest_bot_audio_id
                    )
                    return EndConversationAction()
            else:
                call_state.is_bot_speaking = True  # type: ignore[attr-defined]
        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: "Websocket", tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        """Create a voice output channel."""
        return StudioVoiceOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    def _start_voice_session(
        self,
        session_id: str,
        sid: str,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ) -> None:
        """Create SocketIO WebSocket Adaptor & start async task for voice streaming."""
        if sid in self.active_connections:
            structlogger.warning(
                "studio_chat._start_voice_session.session_already_active",
                session_id=sid,
            )
            return

        structlogger.info(
            "studio_chat._start_voice_session.starting_session", session_id=sid
        )

        # Create a websocket adapter for this connection
        ws_adapter = SocketIOVoiceWebsocketAdapter(
            sio=self.sio,
            session_id=session_id,
            sid=sid,
            bot_message_evt=self.bot_message_evt,
        )
        self.active_connections[sid] = ws_adapter

        # Start voice streaming in an async task
        task = asyncio.create_task(
            self._handle_voice_streaming(on_new_message, ws_adapter, sid)
        )
        self.background_tasks[sid] = task
        task.add_done_callback(lambda _: self._cleanup_tasks_for_sid(sid))

    async def _handle_voice_streaming(
        self,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        ws_adapter: "Websocket",
        sid: str,
    ) -> None:
        """Handle voice streaming for a Socket.IO connection."""
        try:
            await self.run_audio_streaming(on_new_message, ws_adapter)
        except Exception as e:
            structlogger.exception(
                "studio_voice.voice_streaming.error",
                error=str(e),
                sid=sid,
            )
            if sid in self.active_connections:
                del self.active_connections[sid]

    def _cleanup_tasks_for_sid(self, sid: str) -> None:
        if sid in self.background_tasks:
            task = self.background_tasks.pop(sid)
            task.cancel()
        if sid in self.active_connections:
            del self.active_connections[sid]

    @hookimpl  # type: ignore[misc]
    def after_server_stop(self) -> None:
        """Cleanup background tasks and active connections when the server stops."""
        structlogger.info("studio_chat.after_server_stop.cleanup")
        self.active_connections.clear()
        for task in self.background_tasks.values():
            task.cancel()

    def blueprint(
        self, on_new_message: Callable[["UserMessage"], Awaitable[Any]]
    ) -> SocketBlueprint:
        socket_blueprint = super().blueprint(
            partial(self.on_message_proxy, on_new_message)
        )

        if not self.sio:
            structlogger.error("studio_chat.blueprint.sio_not_initialized")
            return socket_blueprint

        @socket_blueprint.listener("after_server_start")  # type: ignore[misc]
        async def after_server_start(
            app: "Sanic", _: asyncio.AbstractEventLoop
        ) -> None:
            self.agent = app.ctx.agent

        @self.sio.on("disconnect", namespace=self.namespace)
        async def disconnect(sid: Text) -> None:
            structlogger.debug("studio_chat.sio.disconnect", sid=sid)
            self._cleanup_tasks_for_sid(sid)

        @self.sio.on("session_request", namespace=self.namespace)
        async def session_request(sid: Text, data: Optional[Dict]) -> None:
            """Overrides the base SocketIOInput session_request handler.

            Args:
              sid: ID of the session (from SocketIO).
              data:
                - session_id: Studio Chat channel is used with a Bridge Architecture
                  (Model Service's Socket Bridge), so we use session_id to remain
                  consistent across the bridge. Session ID becomes the sender_id
                  for the UserMessage.
                - is_voice: Boolean indicating if its a voice session.
            """
            # Call parent session_request handler first
            await self.handle_session_request(sid, data)

            # start a voice session if requested
            if data and data.get("is_voice", False):
                self._start_voice_session(data["session_id"], sid, on_new_message)

        @self.sio.on(self.user_message_evt, namespace=self.namespace)
        async def handle_message(sid: Text, data: Dict) -> None:
            """Overrides the base SocketIOInput handle_message handler."""
            # Handle voice messages
            if "audio" in data or "marker" in data:
                if sid in self.active_connections:
                    # Route audio messages to the voice adapter queue
                    ws = self.active_connections[sid]
                    ws.put_message(data)
                return

            try:
                # Handle text messages
                await self.handle_user_message(sid, data, on_new_message)
            except Exception as e:
                structlogger.exception(
                    "studio_chat.sio.handle_message.error",
                    error=str(e),
                    sid=sid,
                )
                await self.sio.emit("error", str(e), room=sid)

        @self.sio.on("update_tracker", namespace=self.namespace)
        async def on_update_tracker(sid: Text, data: Dict) -> None:
            await self.handle_tracker_update(sid, data)

        return socket_blueprint


class StudioVoiceOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "studio_chat"

    def rasa_audio_bytes_to_channel_bytes(
        self, rasa_audio_bytes: RasaAudioBytes
    ) -> bytes:
        return audioop.ulaw2lin(rasa_audio_bytes, 4)

    def channel_bytes_to_message(self, recipient_id: str, channel_bytes: bytes) -> str:
        return json.dumps({"audio": base64.b64encode(channel_bytes).decode("utf-8")})

    def create_marker_message(self, recipient_id: str) -> Tuple[str, str]:
        message_id = uuid.uuid4().hex
        return json.dumps({"marker": message_id}), message_id


class SocketIOVoiceWebsocketAdapter:
    """Adapter to make Socket.IO work like a Sanic WebSocket for voice channels."""

    def __init__(
        self, sio: "AsyncServer", session_id: str, sid: str, bot_message_evt: str
    ) -> None:
        self.sio = sio
        self.bot_message_evt = bot_message_evt
        self._closed = False
        self._receive_queue: asyncio.Queue[Any] = asyncio.Queue()

        # the messages need to be emitted on room=sid
        self.sid = sid

        # used by collect_call_parameters
        # ultimately, this becomes the sender_id
        self.session_id = session_id

    @property
    def closed(self) -> bool:
        return self._closed

    async def send(self, data: Any) -> None:
        """Send data to the client."""
        if not self.closed:
            await self.sio.emit(self.bot_message_evt, data, room=self.sid)

    async def recv(self) -> Any:
        """Receive data from the client."""
        if self.closed:
            raise ConnectionError("WebSocket is closed")
        return await self._receive_queue.get()

    def put_message(self, message: Any) -> None:
        """Put message in the internal receive queue."""
        self._receive_queue.put_nowait(message)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        """Close the connection."""
        self._closed = True
        # at this point, the client should have disconnected

    def __aiter__(self) -> "SocketIOVoiceWebsocketAdapter":
        """Allow the adapter to be used in an async for loop."""
        return self

    async def __anext__(self) -> Any:
        if self.closed:
            raise StopAsyncIteration
        try:
            message = await self.recv()
            return message
        except Exception:
            raise StopAsyncIteration



from __future__ import annotations

import logging
import typing
import warnings
from typing import Any, Dict, List, Optional, Text, Tuple, Type

import numpy as np

import rasa.shared.utils.io
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.nlu.classifiers import LABEL_RANKING_LENGTH
from rasa.nlu.classifiers.classifier import IntentClassifier
from rasa.nlu.featurizers.dense_featurizer.dense_featurizer import DenseFeaturizer
from rasa.shared.constants import DOCS_URL_TRAINING_DATA_NLU
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import TEXT
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.utils.tensorflow.constants import FEATURIZERS

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    import sklearn


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.INTENT_CLASSIFIER, is_trainable=True
)
class SklearnIntentClassifier(GraphComponent, IntentClassifier):
    """Intent classifier using the sklearn framework."""

    @classmethod
    def required_components(cls) -> List[Type]:
        """Components that should be included in the pipeline before this component."""
        return [DenseFeaturizer]

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            # C parameter of the svm - cross validation will select the best value
            "C": [1, 2, 5, 10, 20, 100],
            # gamma parameter of the svm
            "gamma": [0.1],
            # the kernels to use for the svm training - cross validation will
            # decide which one of them performs best
            "kernels": ["linear"],
            # We try to find a good number of cross folds to use during
            # intent training, this specifies the max number of folds
            "max_cross_validation_folds": 5,
            # Scoring function used for evaluating the hyper parameters
            # This can be a name or a function (cfr GridSearchCV doc for more info)
            "scoring_function": "f1_weighted",
            "num_threads": 1,
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        clf: Optional["sklearn.model_selection.GridSearchCV"] = None,
        le: Optional["sklearn.preprocessing.LabelEncoder"] = None,
    ) -> None:
        """Construct a new intent classifier using the sklearn framework."""
        from sklearn.preprocessing import LabelEncoder

        self.component_config = config
        self._model_storage = model_storage
        self._resource = resource

        if le is not None:
            self.le = le
        else:
            self.le = LabelEncoder()
        self.clf = clf

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> SklearnIntentClassifier:
        """Creates a new untrained component (see parent class for full docstring)."""
        return cls(config, model_storage, resource)

    @staticmethod
    def required_packages() -> List[Text]:
        """Any extra python dependencies required for this component to run."""
        return ["sklearn"]

    def transform_labels_str2num(self, labels: List[Text]) -> np.ndarray:
        """Transforms a list of strings into numeric label representation.

        :param labels: List of labels to convert to numeric representation
        """
        return self.le.fit_transform(labels)

    def transform_labels_num2str(self, y: np.ndarray) -> np.ndarray:
        """Transforms a list of strings into numeric label representation.

        :param y: List of labels to convert to numeric representation
        """
        return self.le.inverse_transform(y)

    def train(self, training_data: TrainingData) -> Resource:
        """Train the intent classifier on a data set."""
        num_threads = self.component_config["num_threads"]

        labels = [e.get("intent") for e in training_data.intent_examples]

        if len(set(labels)) < 2:
            rasa.shared.utils.io.raise_warning(
                "Can not train an intent classifier as there are not "
                "enough intents. Need at least 2 different intents. "
                "Skipping training of intent classifier.",
                docs=DOCS_URL_TRAINING_DATA_NLU,
            )
            return self._resource

        y = self.transform_labels_str2num(labels)
        training_examples = [
            message
            for message in training_data.intent_examples
            if message.features_present(
                attribute=TEXT, featurizers=self.component_config.get(FEATURIZERS)
            )
        ]
        X = np.stack(
            [self._get_sentence_features(example) for example in training_examples]
        )
        # reduce dimensionality
        X = np.reshape(X, (len(X), -1))

        self.clf = self._create_classifier(num_threads, y)

        with warnings.catch_warnings():
            # sklearn raises lots of
            # "UndefinedMetricWarning: F - score is ill - defined"
            # if there are few intent examples, this is needed to prevent it
            warnings.simplefilter("ignore")
            self.clf.fit(X, y)

        self.persist()
        return self._resource

    @staticmethod
    def _get_sentence_features(message: Message) -> np.ndarray:
        _, sentence_features = message.get_dense_features(TEXT)
        if sentence_features is not None:
            return sentence_features.features[0]

        raise ValueError(
            "No sentence features present. Not able to train sklearn policy."
        )

    def _num_cv_splits(self, y: np.ndarray) -> int:
        folds = self.component_config["max_cross_validation_folds"]
        return max(2, min(folds, np.min(np.bincount(y)) // 5))

    def _create_classifier(
        self, num_threads: int, y: np.ndarray
    ) -> "sklearn.model_selection.GridSearchCV":
        from sklearn.model_selection import GridSearchCV
        from sklearn.svm import SVC

        C = self.component_config["C"]
        kernels = self.component_config["kernels"]
        gamma = self.component_config["gamma"]
        # dirty str fix because sklearn is expecting
        # str not instance of basestr...
        tuned_parameters = [
            {"C": C, "gamma": gamma, "kernel": [str(k) for k in kernels]}
        ]

        # aim for 5 examples in each fold

        cv_splits = self._num_cv_splits(y)

        return GridSearchCV(
            SVC(C=1, probability=True, class_weight="balanced"),
            param_grid=tuned_parameters,
            n_jobs=num_threads,
            cv=cv_splits,
            scoring=self.component_config["scoring_function"],
            verbose=1,
        )

    def process(self, messages: List[Message]) -> List[Message]:
        """Return the most likely intent and its probability for a message."""
        for message in messages:
            if self.clf is None or not message.features_present(
                attribute=TEXT, featurizers=self.component_config.get(FEATURIZERS)
            ):
                # component is either not trained or didn't
                # receive enough training data or the input doesn't
                # have required features.
                intent = None
                intent_ranking = []
            else:
                X = self._get_sentence_features(message).reshape(1, -1)

                intent_ids, probabilities = self.predict(X)
                intents = self.transform_labels_num2str(np.ravel(intent_ids))
                # `predict` returns a matrix as it is supposed
                # to work for multiple examples as well, hence we need to flatten
                probabilities = probabilities.flatten()

                if intents.size > 0 and probabilities.size > 0:
                    ranking = list(zip(list(intents), list(probabilities)))[
                        :LABEL_RANKING_LENGTH
                    ]

                    intent = {"name": intents[0], "confidence": probabilities[0]}

                    intent_ranking = [
                        {"name": intent_name, "confidence": score}
                        for intent_name, score in ranking
                    ]
                else:
                    intent = {"name": None, "confidence": 0.0}
                    intent_ranking = []

            message.set("intent", intent, add_to_output=True)
            message.set("intent_ranking", intent_ranking, add_to_output=True)

        return messages

    def predict_prob(self, X: np.ndarray) -> np.ndarray:
        """Given a bow vector of an input text, predict the intent label.

        Return probabilities for all labels.

        :param X: bow of input text
        :return: vector of probabilities containing one entry for each label.
        """
        if self.clf is None:
            raise RasaException(
                "Sklearn intent classifier has not been initialised and trained."
            )

        return self.clf.predict_proba(X)

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Given a bow vector of an input text, predict most probable label.

        Return only the most likely label.

        :param X: bow of input text
        :return: tuple of first, the most probable label and second,
                 its probability.
        """
        pred_result = self.predict_prob(X)
        # sort the probabilities retrieving the indices of
        # the elements in sorted order

        sorted_indices = np.fliplr(np.argsort(pred_result, axis=1))
        return sorted_indices, pred_result[:, sorted_indices]

    def persist(self) -> None:
        """Persist this model into the passed directory."""
        import skops.io as sio

        with self._model_storage.write_to(self._resource) as model_dir:
            file_name = self.__class__.__name__
            classifier_file_name = model_dir / f"{file_name}_classifier.skops"
            encoder_file_name = model_dir / f"{file_name}_encoder.json"

            if self.clf and self.le:
                # convert self.le.classes_ (numpy array of strings) to a list in order
                # to use json dump
                rasa.shared.utils.io.dump_obj_as_json_to_file(
                    encoder_file_name, list(self.le.classes_)
                )
                sio.dump(self.clf.best_estimator_, classifier_file_name)

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> SklearnIntentClassifier:
        """Loads trained component (see parent class for full docstring)."""
        import skops.io as sio
        from sklearn.preprocessing import LabelEncoder

        try:
            with model_storage.read_from(resource) as model_dir:
                file_name = cls.__name__
                classifier_file = model_dir / f"{file_name}_classifier.skops"

                if classifier_file.exists():
                    unknown_types = sio.get_untrusted_types(file=classifier_file)

                    if unknown_types:
                        logger.error(
                            f"Untrusted types ({unknown_types}) found when "
                            f"loading {classifier_file}!"
                        )
                        raise ValueError()
                    else:
                        classifier = sio.load(classifier_file, trusted=unknown_types)

                    encoder_file = model_dir / f"{file_name}_encoder.json"
                    classes = rasa.shared.utils.io.read_json_file(encoder_file)

                    encoder = LabelEncoder()
                    intent_classifier = cls(
                        config, model_storage, resource, classifier, encoder
                    )
                    # convert list of strings (class labels) back to numpy array of
                    # strings
                    intent_classifier.transform_labels_str2num(classes)
                    return intent_classifier
        except ValueError:
            logger.debug(
                f"Failed to load '{cls.__name__}' from model storage. Resource "
                f"'{resource.name}' doesn't exist."
            )
        return cls(config, model_storage, resource)



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionUpdateCardStatus(Action):
    def name(self) -> str:
        return "action_update_card_status"

    def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[str, Any]
    ) -> List[Dict[Text, Any]]:
        username = tracker.get_slot("username")
        card_number = tracker.get_slot("card_number")
        new_status = "inactive"

        db = Database()

        # Get user information
        user = db.get_user_by_name(username)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Get card information to verify it belongs to the user
        card = db.get_card_by_number(card_number)
        if not card or card["user_id"] != user["id"]:
            dispatcher.utter_message(text="Card not found or does not belong to you.")
            return []

        # Update card status
        success = db.update_card_status(card_number, new_status)

        if success:
            dispatcher.utter_message(text=f"Card status updated to {new_status}.")
            return [SlotSet("card_status", new_status)]
        else:
            dispatcher.utter_message(
                text="Failed to update card status. Please try again."
            )
            return []
